#!/usr/bin/env python
'''
    Create IMa input file from four-gamete filtered VCF files.

    An IM analysis requires an IM-formatted input file that contains multiple 
    loci, where each locus has sequence information for variant sites as 
    determined by multiple individuals (which are grouped by population).
    To produce this file, this script takes either a VCF file with a
    BED file that indicates loci to be used or a VCF file per locus. A model
    file will be used to split the samples in the VCF(s) into populations.

    For each locus, a header line is created that contains several pieces of
    information, including number of individuals per population at this locus,
    number of sites in sequences provided, mutation model, inheritance scalar,
    and mutation rate per year at locus (over locus, not per base pair). 
    Population sizes and sequence ordering are handled internally, the 
    inheritance scalar and mutation rate can be set via commandline.

    Loci provided as input must pass the four-gamete filtering criteria. 
    In addition, if filtering has not been previously done options are
    available to filter out indels (default), multiallelic sites, and 
    CpGs (with reference genome). Sites with missing data can either be
    filtered by dropping individuals missing data from a locus from 
    analysis at that locus, or by replacing the missing site with a 
    reference allele. 

    ###############
    Input Arguments
    ###############
    **--vcf** *<input_filename>*
        Filename for input VCF if using BED file with locus information
    **--vcfs** *<vcf_filename_1>*...*<vcf_filename_n>*
        One or multiple VCF input filenames where each file contains
        sequences for a single locus. A file with lines corresponding
        to filenames can be provided with **--vcfs** *@<vcf_filelist>*
    **--model-file** *<model_filename>*
        Filename of model file.
    **--model** *<model name>*
        If model file contains multiple models, use this argument
        to specify name of population to use.
    **--reference-fasta** *<reference_filename>*
        Filename for reference FASTA file. File can be uncompressed or
        bg-zipped, but must be indexed with faidx. When option is specified,
        default options are to include sequence in output loci but not filter
        for CpGs (use --parse-cpg)
    **--bed** *<bed_filename>*
        Filename for BED file specifying loci if only one VCF is provided. 
        Can be used with multiple VCFs if line count aligns, used for getting
        correct locus length.
    
    ################
    Output Arguments
    ################
    **--out** *<out_filename>*
        Output filename.
    
    #############
    Model Options
    #############
    **--mutrate** *<mutation rate>*
        Set mutation rate per base pair (default is 1e-9). This value is
        multiplied by locus length to get mutation rate per locus.
    **--inheritance-scalar** *<scalar>*
        Sets inheritance scalar for all loci. Default behavior is to 
        set scalar to 1 for non-X/Y/MT chromosomes, .75 for 'X' and
        'chrX', and .25 for 'y', 'chrY', 'MT', and 'chrMT'.

    #################
    Filtering Options
    #################
    **--remove-multiallele**
        Set all multiallelic sites to be reference. 
    **--drop-missing-sites** *<individual_count>*
        Drops all sites where more than 'individual_count' individuals
        are missing data. Default is -1 (no dropping), and 0 will 
        drop all sites missing data and replace them with the reference
        allele. 
    **--drop-missing-inds**
        If set, if an individual is missing data at a locus, that 
        individual will not be included at that locus and population
        counts for that locus will be adjusted.
    **--remove-cpg**
        Requires --reference-fasta. If set, will replace CpG sites with
        reference allele at site, setting them as invariant.

    #############
    Other Options
    #############
    **--zero-ho**
        Use if input BED file uses zero-based, half-open coordinates instead
        of one-based, closed.
    **--bed-column-index** *<start_col,end_col,chrom_col>*
        Comma-separated list of zero-based indexes of start, end, and 
        chromosome name columns in input BED file. Default value for
        traditionally structured BED is 1,2,0
    **--noseq**
        If set, and --reference-fasta is provided, will not output 
        invariant sites to IM file.

'''
import sys
import pysam
import argparse
import os
import logging
from random import sample
from collections import OrderedDict

import pgpipe.vcf_reader_func as vf
from pgpipe.logging_module import initLogger, logArgs
from pgpipe.genome_region import Region, RegionList
from pgpipe.parse_functions import defaultsDictForFunction, getConfigFilename, makeRequiredList, getArgsWithConfig
from pgpipe.model import Model, read_single_model
#from tabix_wrapper import prepVcf

#Input: VCF file, reference sequence, region list (possibly .bed file)
#Output: Sequences with reference genome overlayed with VCF SNP calls

#sys.path.insert(0,os.path.abspath(os.path.join(os.pardir, 'andrew')))


class locus():
    #Name, pops, store for seqs, length, mut model, scalar, mutrate
    def __init__(self, gener, rec_list, popmodel, args):
        if args.inhet_sc is None:
            if gener.chrom in ['X','chrX']:
                self.inhet_sc = 0.75
            elif gener.chrom in ['Y','chrY','MT','chrMT']:
                self.inhet_sc = 0.25
            else:
                self.inhet_sc = 1
        else:
            self.inhet_sc = args.inhet_sc
        self.name = gener.chrom+':'+str(gener.start)+':'+str(gener.end)
        self.gene_len = gener.end - gener.start
        for rec in rec_list:
            self.gene_len += (getMaxAlleleLength(rec.alleles) - len(rec.ref))
        self.popmodel = popmodel
        self.popkeys = []
        self.seqs = {}
        for p in popmodel.pop_list:
            self.popkeys.append(p)
            self.seqs[p] = []
        self.mut_model = 'I'
        #if args.mut_model is not None:
        #    if args.mut_model not in ['I','H','S','J','IS']:
        #        raise Exception("%s is an invalid mutation model (must select from I,H,S,J,IS)" % (args.mut_model))
        #    self.mut_model = args.mut_model
        if args.refname is not None and args.printseq:
            self.seq_len = self.gene_len
        else:
            self.seq_len = len(rec_list)
        self.mutrate = self.gene_len * args.mutrate

    def addSeq(self,seq,pop):
        try:
            self.seqs[pop].append(seq)
        except IndexError:
            raise Exception("Pop %s is not in pop list" % pop)

    def printToFile(self, outf = sys.stdout):
        out_str = ''
        out_str += (self.name)+' '
        for p in self.popkeys:
            out_str += (str(len(self.seqs[p]))+' ')
        out_str += str(self.seq_len)+' '
        out_str += self.mut_model + ' '
        out_str += str(self.inhet_sc)+' '
        out_str += str("%.14f" % (self.mutrate))+'\n'
        outf.write(out_str)
        for p in self.popkeys:
            for seq in self.seqs[p]:
                outf.write(seq+'\n')

class outputBuffer():

    def __init__(self, popmodel, outfile=sys.stdout):
        self.header = None
        self.hold = True
        self.outfile = outfile
        self.popmodel = popmodel
        self.poptree = None
        self.loci = []

    def createHeader(self):
        self.header = {}
        self.header['title'] = 'Test IMa input'
        self.header['pops'] = [p for p in self.popmodel.pop_list]

    def writeHeader(self):
        self.outfile.write(self.header['title']+'\n')
        self.outfile.write(str(len(self.header['pops']))+'\n')
        self.outfile.write(' '.join(self.header['pops'])+'\n')
        if self.poptree is not None:
            self.outfile.write(self.poptree+'\n')
        self.outfile.write(str(len(self.loci)+'\n'))



def createParser():
    parser = argparse.ArgumentParser(description=("Generates an IMa input "
                                     "file from a VCF file, a reference"
                                     " genome, a list of gene regions, "
                                     "and a population info file."),
                                     fromfile_prefix_chars="@")
    parser.add_argument("--vcf", dest="vcfname", help="Input VCF filename")
    parser.add_argument("--vcfs", dest="vcflist", metavar="VCF",nargs="+",
                        help=("Input VCF per loci"))
    parser.add_argument("--reference-fasta", dest="refname", help="Reference FASTA file, use this plus --remove-cpg to remove CpGs from loci")
    parser.add_argument("--bed", dest="genename", help=("If using a single "
                        "VCF, list of regions to generate loci from"))
    parser.add_argument("--bed-column-index", dest="gene_col",
                        help=("Comma-separated list of length 3 with 0-based"
                        " indexes of start, end, and chromosome data in input"
                        " BED file. Default for normal BED is 1,2,0"))
    parser.add_argument("--model-file", dest="popname", help=("Filename of "
                        "pop model file"))
    parser.add_argument("--model",dest="poptag",help=("If model file has "
                        "multiple models, use model with this name"))
    parser.add_argument("--zero-ho", dest="zeroho", action="store_true",
                        help="Region coordinates are zero-based, half-open")
    parser.add_argument("--zero-closed", dest="zeroclosed",action="store_true",
                        help="Region coordinates are zero-based, closed")
    parser.add_argument("--keep-indels", dest="indel_flag", action="store_true",
                        help="Include indels when reporting sequences")
    parser.add_argument("--remove-multiallele",dest="remove_multiallele",
                        action="store_true", help=("Remove sites with more "
                        "than two alleles"))
    #Need to figure out exactly how this option is working
    parser.add_argument("--drop-missing-sites",dest="remove_missing",
                        action="store_const",const=0,default=-1,help=("Will "
                        "use reference allele for sites where more than n "
                        "individuals are missing data. Missing data will "
                        "by default cause an error."))
    parser.add_argument("--drop-missing-inds",dest="drop_inds",action="store_true",
                        help="Drop individuals from loci if they are missing any data")
    parser.add_argument("--remove-cpg",dest="parsecpg",action="store_true",
                        help=("Will replace CpG sites with reference allele"
                        ", requires --reference-fasta"))
    parser.add_argument("--noseq",dest="printseq",action="store_false",
                        help=("Will only print variants when reference is "
                        "provided. Used so CpG filtering can be done if "
                        "invariant sites aren't desired in output"))
    parser.add_argument("--trim-to-ref-length", dest="trim_seq",
                        action="store_true",
                        help=("Trims sequences if indels cause them to be "
                        "longer than reference"))
    parser.add_argument("--out", dest="output_name", default="input.ima.u",
                        help= ("Optional name for output other than default"))
    parser.add_argument("--compress-vcf", dest="compress_flag",
                        action="store_true", help=("If input VCF is not "
                        "compressed, will compress and use zip search"))
    parser.add_argument("--conf", dest="config_name", help= ("Name of "
                        "file with configuration options"))
    parser.add_argument("--no-ref-check", dest="ref_check",
                        action="store_false", help=("Prevents exception "
                        "generated by mismatched reference alleles from "
                        "VCF file compared to reference"))
    parser.add_argument("--mutrate",dest="mutrate",type=float,default=1e-9,
                        help="Mutation rate per base pair (default is 1e-9)")
    parser.add_argument("--inheritance-scalar",dest="inhet_sc",type=float,
                        default=None,help=("Sets inheritance scalar for all "
                        "chromosomes (default is 1 for autosomes, .75 for X,"
                        " .25 for Y/MT)"))
    parser.add_argument("--output-fasta",dest="fasta",action="store_true",
                        help=("Output fasta file(s) instead of IMa input. "
                        "Only vaguely supported."))
    parser.add_argument("--out-prefix",dest="multi_out",type=str,
                        help=("If output is fasta, generate one file"
                        "per loci."))
    return parser


def checkArgs(args):
    if args.vcfname is None and args.vcflist is None:
        raise Exception("Must provide at least one VCF file to either --vcf or --vcfs")
    if args.vcfname is not None and args.vcflist is not None:
        raise Exception("Cannot use both arguments --vcf and --vcfs")
    if args.vcfname is not None and args.genename is None:
        raise Exception("If using --vcf, must provide a BED file with --bed")
    if args.refname is None and args.parsecpg:
        raise Exception("CpG parsing requires reference genome (--reference-fasta)")
    if args.popname is None and not args.fasta:
        raise Exception("IMa file requires model input with --model-file [filename]")


def validateFiles(args):
    """Validates that files provided to args all exist on users system"""
    for var in ['vcfname', 'refname', 'genename','popname']:
        f = vars(args)[var]
        if f is not None and not os.path.exists(f):
            raise ValueError('Filepath for %s not found at %s' %
                            (var, f))
    if args.vcflist is not None:
        for f in args.vcflist:
            if not os.path.exists(f):
                raise ValueError(('File not found at %s') % f)


def getMaxAlleleLength(alleles):
    """If an indel, returns length of longest allele (returns 1 for snp)"""
    return max([len(r) for r in alleles])


def checkRefAlign(vcf_recs, fasta_ref, chrom, ref_check):
    """Compares sequence from record to reference FASTA sequence"""
    for vcf_r in vcf_recs:
        vcf_seq = vcf_r.ref
        pos = vcf_r.pos-1
        try:
            fasta_seq = fasta_ref.fetch(vcf_r.chrom, pos, pos+len(vcf_seq))
        except KeyError:
            fasta_seq = fasta_ref.fetch(vf.flipChrom(vcf_r.chrom),pos,pos+len(vcf_seq))
        if vcf_seq.upper() != fasta_seq.upper():
            if ref_check:
                raise Exception(("VCF bases and reference bases do not match."
                        "\nVCF reference: %s\nFASTA reference: "
                        "%s\nPosition: %s") % (vcf_seq, fasta_seq, str(pos)))
            else:
                logging.warning("Bases at position %s do not match" %
                            str(pos))
                break


def generateSequence(rec_list, ref_seq, region, chrom, indiv, idx, args):
    """Fetches variant sites from a given region, then outputs sequences
    from each individual with their correct variants. Will print sequence
    up to the variant site, then print correct variant. After last site,
    will output the rest of the reference sequence."""
    #var_sites = vcf_reader.fetch(chrom,region.start,region.end)
    fl = 0
    seq = ''
    prev_offset = 0

    for vcf_record in rec_list:
        issnp = vf.checkRecordIsSnp(vcf_record)
        if not args.indel_flag and not issnp:
            continue

        #If reference included, add all bases between last and current variant
        pos_offset = vcf_record.pos - 1 - region.start
        if ref_seq is not None:
            for i in range(prev_offset, pos_offset):
                seq += ref_seq[i]

        allele = vcf_record.samples[indiv].alleles[idx]
        if allele is None:
            raise Exception(("Individual %d at position %d is missing "
            "data") % (vcf_record.pos,indiv))
        if issnp:
            #Place allele, move reference offset
            seq += vcf_record.samples[indiv].alleles[idx]
            prev_offset = pos_offset+1
        else:
            #Find longest allele, pad others to its length
            #Offset by length of reference allele
            max_indel = getMaxAlleleLength(vcf_record.alleles)
            allele = vcf_record.samples[indiv].alleles[idx]
            if allele is None:
                allele = 'N'
            for i in range(len(allele), max_indel):
                allele += '_'
            seq += allele
            indel_offset = len(vcf_record.ref)
            prev_offset = pos_offset+indel_offset
    #Output remaining sequence
    if ref_seq is not None:
        for i in range(prev_offset, len(ref_seq)):
            seq += ref_seq[i]
        if args.trim_seq:
            return seq[:len(ref_seq)]
    return seq



def writeHeader(popmodel, loci_count, out_f, mutrate, header="Test IMa input",
                pop_string=None):
    out_f.write(header+'\n')
    out_f.write('#Mutation rate: '+str(mutrate)+'\n')
    out_f.write(str(popmodel.npop)+'\n')
    pops = ''
    for p in popmodel.pop_list:
        pops += (p+' ')
    out_f.write(pops+'\n')
    #Add default treestring when found in IMa3 source
    if pop_string is not None:
        out_f.write(pop_string+'\n')
    out_f.write(str(loci_count)+'\n')

def getLocusHeader(gener, popmodel, rec_list, mut_model="I", inhet_sc=None, mut_rate=1e-9,include_seq=False):
    if inhet_sc is None:
        if gener.chrom in ['X','chrX']:
            inhet_sc = 0.75
        elif gener.chrom in ['Y','chrY','MT','chrMT']:
            inhet_sc = 0.25
        else:
            inhet_sc = 1
    name = gener.chrom+':'+str(gener.start)+':'+str(gener.end)
    gene_len = gener.end-gener.start
    for rec in rec_list:
        gene_len += (getMaxAlleleLength(rec.alleles) - len(rec.ref))
    lh = name
    #for i in range(len(pop_data)):
    #    lh += ' '+str(len(pop_data[i][1]))
    for p in popmodel.pop_list:
        lh += ' '+str(2*popmodel.nind[p])
    if include_seq:
        lh += ' '+str(gene_len)
    else:
        lh += ' '+str(len(rec_list))
    if mut_model not in ['I','H','S','J','IS']:
        raise Exception("%s is an invalid mutation model (must select from I,H,S,J,IS)" % (mut_model))
    lh += ' '+mut_model
    lh += ' '+str(inhet_sc)
    mutlocus = mut_rate * gene_len
    lh += ' '+str("%.14f" % (mutlocus))
    return lh

def getMultiName(args):
    suffix = '.fasta' if args.fasta else '.u'
    i = 1
    while True:
        yield args.multi_out+'_region'+str(i)+suffix
        i += 1

def getOutputFilename(args):
    if args.output_name is not None:
        return args.output_name
    suffix = '.fasta' if args.fasta else '.u'
    va = args.vcfname.split('.')
    if len(va) == 1:
        return va+suffix
    ext_cut = -1
    if va[-1] == 'gz':
        ext_cut = -2
    return '.'.join(va[:ext_cut])+suffix

def hasMissingData(rec_list, indiv_idx):
    for rec in rec_list:
        if rec.samples[indiv_idx].alleles[0] is None:
            return True
    return False


def vcf_to_ima(sys_args):
    """Returns an IMa input file given four-gamete filtered loci in one or 
    multiple VCFs.

    An IM analysis requires a file with variants for a set of individuals at
    multiple loci. These loci must be checked so that there are no 
    recombination events between SNPs, as determined by a four-gamete test.
    Variant input can be either a single VCF with a BED file listing regions
    that correspond to loci, or a single VCF per locus. Also required is a
    model file that lists what individuals in the VCF header correspond to
    which population. Filtering can be done in this step, for indels,
    multiallelic variants, sites with missing data, and CpGs (reference
    required). Mutation rate and inheritance scalars can also be set via
    commandline. 

    Parameters
    ----------
    --vcf : str
        Filename for VCF input file. 
    --vcfs : str (1+)
        Filenames for VCF locus files. Can also call '--vcfs @[filename.txt]'
        where filename.txt has paths for all desired VCF files.
    --reference-fasta : str
        Filename for FASTA reference file. File must contain entirety of 
        chromosomes specified. By default this will put invaraint sites 
        in the output loci. Required for CpG filtering.
    --model-file : str
        Filename for model file. This file contains population designations
        for individuals to be used in IM analysis. 
    --model : str
        If model file has more than one population, use model with this name
    --bed : str
        Filename for loci file with single VCF input. Requires columns with
        start, end, and chromosome of each locus.
    --bed-column-index : ints
        Comma-separated list of length 3 with 0-based indexes of start, end,
        and chromosome data in input BED file. Default for normal BED is
        "1,2,0".
    --zero-ho : bool
        If set, treats regions in BED file as 0-based, (h)alf (o)pen
        coordinates rather than 1-based, closed. 
    --out : str
        Name for output file. Defaults to 'input.ima.u"
    --remove-multiallele : bool
        If set, will set all multiallelic sites found as reference allele
    --drop-missing-inds : bool
        If set, if an individual at a locus is missing data, that individual
        will not be included at the locus.
    --remove-cpg : bool
        If set, will replace CpG sites with reference allele. Requires
        --reference-fasta.
    --noseq : bool
        If set and --reference-fasta is called, will only use for CpG
        checking and not output invariant sites in output file.



    Other Parameters
    ----------------
    --remove-multiallele : bool
        If set, will set all multiallelic sites found as reference allele
    --drop-missing-inds : bool
        If set, if an individual at a locus is missing data, that individual
        will not be included at the locus.
    --remove-cpg : bool
        If set, will replace CpG sites with reference allele. Requires
        --reference-fasta.
    --noseq : bool
        If set and --reference-fasta is called, will only use for CpG
        checking and not output invariant sites in output file.
    --trim-to-ref-length: bool, optional (False)
        If set, the sequences output will always match the length of the
        region they are found in. For example, a sequence with an insertion
        will cause the sequence to be an additional length of n-1, with n
        being the length of the insertion.
    --mutrate : float
        Use custom mutation rate for file (default is 1e-9)
    --inheritance-scalar : float
        Use fixed inheritance scalar for every output loci. Default is
        1 for autosomal chromosomes, .75 for X and .25 for Y/MT.


    Output
    ------
    IMa file
        Will be named either '--out' value or (vcfinput).ima.u.
        Contains variants in designated loci for IMa run.
    """
    parser = createParser()
    if len(sys_args) == 0:
        parser.print_help()
        sys.exit(1)

    required_args = ['vcfname','popname']
    args = getArgsWithConfig(parser,sys_args,required_args,'vcf_to_ima')
    checkArgs(args)
    logArgs(args)
    #validateFiles(args)
    if args.multi_out is None:
        output_filename = getOutputFilename(args)
        output_file = open(output_filename, 'w')
    else:
        outnamegen = getMultiName(args)
    popmodel = None
    use_allpop = False
    if args.popname is not None:
        popmodel = read_single_model(args.popname,args.poptag)
    else:
        use_allpop = True

    filter_recs = (args.remove_multiallele or (args.remove_missing != -1) or not args.indel_flag or args.parsecpg)

    if args.vcfname is not None:
        single_file = True
    elif args.vcflist is not None:
        single_file = False
    else:
        raise Exception("VCF tag must be specified")
    if single_file:
        vn = args.vcfname
    else:
        vn = args.vcflist[0]
    vcf_reader = vf.VcfReader(vn,
                              compress_flag=args.compress_flag,
                              popmodel=popmodel,
                              use_allpop=use_allpop)
    logging.info('VCF file read')

    regions_provided = False
    if args.genename is not None:
        regions_provided = True
        region_list = RegionList(filename=args.genename, zeroho=args.zeroho,
                            zeroclosed=args.zeroclosed,
                            colstr=args.gene_col)
        logging.info('Region list read')
    fasta_ref = None
    if args.refname is not None:
        fasta_ref = pysam.FastaFile(args.refname)
    record_count = 1
    first_el = vcf_reader.info_rec

    logging.info('Total individuals: %d' % (len(vcf_reader.info_rec.samples)))
    if regions_provided:
        logging.info('Total regions: %d' % (len(region_list.regions)))
    total_regions = (len(region_list.regions) if regions_provided else len(args.vcflist))
    if not args.fasta:
        writeHeader(popmodel, total_regions, output_file, args.mutrate)
    if not single_file:
        vcf_reader.reader.close()
    for i in range(total_regions):
        #if regions_provided:
        #region = region_list.regions[i]
        if args.multi_out is not None:
            try:
                output_file.close()
            except:
                pass
            output_filename = next(outnamegen)
            output_file = open(output_filename,'w')
        if single_file:
            region = region_list.regions[i]
            rec_list = vcf_reader.getRecordList(region)
        else:
            vcf_reader = vf.VcfReader(args.vcflist[i],
                                      compress_flag=args.compress_flag,
                                      popmodel=popmodel,
                                      use_allpop=use_allpop)
            rec_list = vcf_reader.getRecordList()
            vcf_reader.reader.close()
            if regions_provided:
                region = region_list.regions[i]
            else:
                region = Region(rec_list[0].pos-1,rec_list[-1].pos,rec_list[0].chrom)
        if filter_recs:
            #Make this modify rec list in place
            t = vf.filterSites(rec_list,remove_cpg=args.parsecpg,remove_indels=(not args.indel_flag),remove_multiallele=args.remove_multiallele,remove_missing=args.remove_missing,inform_level=1,fasta_ref=fasta_ref)
            rec_list = t
        if len(rec_list) == 0:
            logging.warning(("Region %s has no variants "
                            "in VCF file") % (region.toStr()))
        logging.debug('Region %d to %d: %d variants' %
                      (region.start,region.end,len(rec_list)))
        ref_seq = None
        if fasta_ref is not None:
            try:
                ref_seq = fasta_ref.fetch(region.chrom, region.start, region.end)
            except KeyError:
                ref_seq = fasta_ref.fetch(vf.flipChrom(region.chrom),region.start,region.end)
            checkRefAlign(rec_list, fasta_ref, region.chrom, args.ref_check)
        if not args.fasta:
            temp_locus = locus(region, rec_list, popmodel, args)
            #reg_header = getLocusHeader(region, popmodel, rec_list,mut_rate=args.mutrate,inhet_sc=args.inhet_sc,include_seq=(fasta_ref is not None))
            #output_file.write(reg_header+'\n')
        popnum = 0
        #for p in popmodel.pop_list:
        for p in vcf_reader.popkeys.keys():
            for indiv,indiv_idx in enumerate(vcf_reader.popkeys[p]):
                if indiv_idx == -1:
                    continue
                for hap in range(len(first_el.samples[indiv_idx].alleles)):
                    if args.fasta:
                        seq = generateSequence(rec_list, ref_seq,
                                   region, region.chrom, indiv_idx, hap, args)
                        output_file.write('>'+first_el.samples[indiv_idx].name+'_'+str(hap)+'\n')
                        output_file.write(seq+'\n')
                        continue
                    hmd = hasMissingData(rec_list,indiv_idx)
                    if hmd:
                        if not args.drop_inds:
                            raise Exception("Individual %d from pop %s at loci %d, site %d is missing data" % (indiv_idx,p,record_count,i))
                        else:
                            continue

                    seq = generateSequence(rec_list, ref_seq,
                                   region, region.chrom, indiv_idx, hap, args)
                    seq_name = str(popnum)+':'+str(indiv)+':'+str(hap)
                    seq_name += ''.join([' ' for i in range(len(seq_name),10)])
                    allseq = seq_name + seq
                    temp_locus.addSeq(allseq,p)
                    #output_file.write(seq_name)

                    #output_file.write(seq+'\n')
                #indiv += 1
            popnum += 1
            indiv = 0
        temp_locus.printToFile(outf = output_file)
        record_count += 1
    output_file.close()

if __name__ == "__main__":
    initLogger()
    vcf_to_ima(sys.argv[1:])



import os
import sys
import logging
import subprocess
import copy

#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir,'pppipe')))

from pgpipe.vcf_reader_func import checkFormat

def stdout_bcftools_reference():

    # Write the log header
    sys.stdout.write('Please Reference alongside the PPP:\n')

    # Write the reference
    sys.stdout.write('Li, H. et al. The Sequence Alignment/Map format and '
                   'SAMtools. Bioinformatics (2009). doi:10.1093/bioinformatics/btp352')

def log_bcftools_reference (out_filename, append_mode = False, ref_header = True):

    # Check if the file is to be written in append mode
    if append_mode:

        # Open the file
        log_file = open(out_filename + '.log', 'a')

        # Check if the ref header should be added
        if ref_header:
            log_file.write('\nPlease Reference alongside the PPP:\n')

    else:

        # Open the file
        log_file = open(out_filename + '.log', 'w')

        # Check if the ref header should be added
        if ref_header:
            log_file.write('Please Reference alongside the PPP:\n')

    log_file.write('Li, H. et al. The Sequence Alignment/Map format and '
                   'SAMtools. Bioinformatics (2009). doi:10.1093/bioinformatics/btp352')
    log_file.close()

    logging.info('Reference assigned')

def assign_position_information (chromosome, from_bp, to_bp):

    # String to store the position information i.e. chr:from_bp-to_bp
    position_information = copy.deepcopy(chromosome)

    # Check if either bp position arguments were specified
    if from_bp or to_bp:

        # List of the position arguments, in their required order
        position_args = [':', from_bp, '-', to_bp]

        # Filter the position arguments to remove empty values
        filttered_position_args = list(filter(None, position_args))

        # Map the arguments to str and add them to the chromosome argument
        position_information += ''.join(map(str, filttered_position_args))

    return position_information

def return_output_format_args (output_format):
    '''
        Return bcftools arguments for output format

        Parameters
        ----------
        output_format : str
            The specified output format

        Raises
        ------
        Exception
            If output format is unsupported by bcftools
    '''

    # Return the output format arguments
    if output_format == 'vcf':
        return ['-O', 'v']
    elif output_format == 'bcf':
        return ['-O', 'b']
    elif output_format == 'vcf.gz':
        return ['-O', 'z']
    else:
        raise Exception('Unsupported file format')

def check_for_bcftools_output (bcftools_output):
    '''
        Checks for the previous bcftools output

        Confirms that neither a previous bcftools log or output file exists.

        Parameters
        ----------
        bcftools_output : str
            Specifies the output filename to be checked

        Raises
        ------
        IOError
            If the bcftools output file exists
        IOError
            If the bcftools log file exists

    '''
    # Check if output file already exists
    if os.path.isfile(bcftools_output):
        raise IOError('Output file already exists. Use --overwrite to ignore')

    # Check if log file already exists
    if os.path.isfile(bcftools_output + '.log'):
        raise IOError('Log file already exists. Use --overwrite to ignore')

def check_bcftools_for_errors (bcftools_stderr):
    '''
        Checks the bgzip stderr for errors

        Parameters
        ----------
        bcftools_stderr : str
            bcftools stderr

        Raises
        ------
        Exception
            If bcftools stderr returns an error
    '''

    # Expand as errors are discovered

    # Log warning messages
    if 'W::' in bcftools_stderr:
        logging.warning(bcftools_stderr.strip())

    # Report errors that are not warnings
    elif bcftools_stderr:
        raise Exception(bcftools_stderr)

def get_unique_chrs (filename):

    # Check if the files are indexed
    if check_for_index(filename) == False:

        # Create an index if not found
        create_index(filename)

    # Get set of the chromosomes
    chromosome_set = pipe_bcftools_to_set(['query', '-f', '%CHROM\n', filename])

    # Return list of the unique chromosomes
    return list(chromosome_set)

def get_samples (filename):

    # Get list of the samples
    sample_list = pipe_bcftools_to_list(['query', '-l', filename])

    # Return list of the samples
    return sample_list

def check_for_index (filename):
    '''
        Checks for index file

        If the file is capable of having an index (i.e. bgzipped-VCF or BCF) the
        function will return either True (i.e. index found) or False. However,
        if the file is a VCF the function will return None (as VCF files cannot
        have an index). An error is returned if the file is either a
        gzipped-VCF file or not a VCF-based format.

        Parameters
        ----------
        filename : str
            Filename of VCF-formatted file

        Returns
        -------
        bool, None
            Returns bool for VCF.GZ and BCF files. Returns None for VCF files

        Raises
        ------
        Exception
            If the file is a gzipped-VCF or of an unknown format
    '''

    # Assign the file format
    file_format = checkFormat(filename)

    # Check if the file to be indexed is a vcf.gz
    if file_format == 'bgzip':

        # Check if the index (.tbi) exists
        if os.path.isfile(filename + '.tbi'):

            # Return True if the file has an index
            return True

    # Check if the file to be indexed is a bcf
    elif file_format == 'bcf':

        # Check if the index (.csi) exists
        if os.path.isfile(filename + '.csi'):

            # Return True if the file has an index
            return True

    # Check if the file is vcf (does not need an index)
    elif file_format == 'vcf':

        # Return None, as uncompressed VCF files do not have an index
        return None

    # Check if the file is gzip-compressed vcf (cannot have an index)
    elif file_format == 'gzip':
        raise Exception('GZIP-compressed VCF files do not support index files.')

    # Check if the file is an unknown format
    else:
        raise Exception('Unknown file format')

    # Return False if the file supports an index, but does not have one
    return False

def delete_index (filename):
    '''
        Deletes an index file

        If the file is capable of having an index (i.e. bgzipped-VCF or BCF)
        this function will delete the index. However, if the file is either a
        VCF or a gzip-compressed VCF the function will return an error. The
        function also results in an error if the index cannot be found. This
        function should be used following check_for_index.

        Parameters
        ----------
        filename : str
            Filename of VCF-formatted file

        Raises
        ------
        Exception
            No index file could be found
        Exception
            If the file is a gzipped-VCF or a VCF
    '''

    # Assign the file format
    file_format = checkFormat(filename)

    # Check if the file to be indexed is a vcf.gz
    if file_format == 'bgzip':
        # Check if the index (.tbi) exists
        if os.path.isfile(filename + '.tbi'):
            # Delete the index
            os.remove(filename + '.tbi')
            return

    # Check if the file to be indexed is a bcf
    elif file_format == 'bcf':
        # Check if the index (.csi) exists
        if os.path.isfile(filename + '.csi'):
            # Delete the index
            os.remove(filename + '.csi')
            return

    # Check if the file is vcf (cannot have an index)
    elif file_format == 'vcf':
        raise Exception('VCF format does not support index files.')

    # Check if the file is gzip-compressed vcf (cannot have an index)
    elif file_format == 'gzip':
        raise Exception('GZIP-compressed VCF files do not support index files.')

    # Return error if no index is found
    raise Exception('No index file found or the index has the incorrect extension.')

def create_index (filename):
    '''
        Creates an index file

        If the file is capable of having an index (i.e. bgzipped-VCF or BCF)
        this function will create an index file. However, if the file is a
        different format the function will return an error.

        Parameters
        ----------
        filename : str
            Filename of VCF-formatted file

        Raises
        ------
        Exception
            If the file is not a bgzipped-VCF or BCF
    '''

    # Assign the file format
    file_format = checkFormat(filename)

    # Check if the file to be indexed is a vcf.gz
    if file_format == 'bgzip':

        # Create a index (.tbi)
        call_bcftools(['index', '-t', filename])

        logging.info('Index file created')

    # Check if the file to be indexed is a bcf
    elif file_format == 'bcf':

        # Create a index (.csi)
        call_bcftools(['index', '-c', filename])

        logging.info('Index file created')

    # Report if file cannot be indexed
    else:
        raise Exception('Error creating index for: %s. Only .bcf and .vcf.gz (bgzip) files are supported.' % filename)

def chr_subset_file (filename, chromosome, output_prefix, output_format, from_bp = None, to_bp = None, filter_args = None, overwrite = False):
    '''
        Creates chromosome subset

        This function is used to create a VCF-formatted subset with only
        the data from a single chromosome.

        Parameters
        ----------
        filename : str
            Filename of VCF-formatted input
        chromosome : str
            Chromosome to subset
        output_prefix : str
            Prefix of the VCF-formatted output (i.e. without file extension)
        output_format : str
            The format of the output (e.g. vcf, bcf, vcf.gz)
        from_bp : int, optional
            Lower bound of sites to include
        to_bp : int, optional
            Upper bound of sites to include
        filter_args : list, optional
            List of arguments to filter the subset file
        overwrite : bool, optional
            Specify if previous output should be overwritten
    '''

    # Assign the output format arguments
    output_args = return_output_format_args(output_format)

    # Stores the specified output filename
    output_path = '%s.%s' % (output_prefix, output_format)

    # Check if previous output should not be overwritten 
    if not overwrite:

        # Check if previous output exists
        if os.path.isfile(output_path):

            raise Exception('Subset file %s already exists' % output_path)

    # Assigns the output file to the arguments
    output_args.extend(['-o', output_path])

    # Creates a list to the arguments and store the bcftools call
    selection_args = ['view']

    # Assigns the input file to the arguments
    selection_args.append(filename)

    # Store the position information i.e. chr:from_bp-to_bp
    position_information = assign_position_information(chromosome, from_bp, to_bp)

    # Checks if the input file has an index
    if check_for_index(filename):

        # Selects the position using the index
        selection_args.extend(['-r', position_information])
    
    else:

        # Selects the position using the stdout stream
        selection_args.extend(['-t', position_information])

    # Check if no filters are required
    if not filter_args:

        # Call bcftools
        call_bcftools(selection_args + output_args)

    else:

        # Pipe bcftools to bcftools
        pipe_bcftools_bcftools(selection_args, filter_args + output_args)
    
def concatenate (filenames, output_prefix, output_format, keep_original = False, optional_args = []):
    '''
        Concatenate multiple VCF-formatted files

        This function will concatenate multiple VCF-formatted files into a
        single VCF-formatted file of the specifed format.

        Parameters
        ----------
        filenames : list
            List of VCF-formatted input filenames
        output_prefix : str
            Prefix of the VCF-formatted output (i.e. without file extension)
        output_format : str
            The format of the output (e.g. vcf, bcf, vcf.gz)
        keep_original : bool, optional 
            Decides if the original files should be kept
        optional_args : list, optional
            List of additional and optional arguments
    '''

    # Holds the arguments to convert to VCF format
    concat_args = ['concat']

    # Assign the output format arguments
    output_format_args = return_output_format_args(output_format)

    # Store the output format arguments
    concat_args.extend(output_format_args)

    # Stores the specified output filename
    vcf_output = '%s.%s' % (output_prefix, output_format)

    # Assigns the output file to the arguments
    concat_args.extend(['-o', vcf_output])

    # Assigns the input files to concatenate
    concat_args.extend(filenames)

    # Call bcftools
    call_bcftools(concat_args + optional_args)

    # Delete the original files once the merged file is created
    if not keep_original:

        # Loop the files
        for filename in filenames:

            # Check if the file has an index
            if check_for_index(filename) == True:

                # Delete the index
                delete_index(filename)

            # Delete the file
            os.remove(filename)

def merge (filenames, output_prefix, output_format, keep_original = False, optional_args = []):
    '''
        Merge multiple VCF-formatted files

        This function will merge multiple VCF-formatted files into a
        single VCF-formatted file of the specifed format.

        Parameters
        ----------
        filenames : list
            List of VCF-formatted input filenames
        output_prefix : str
            Prefix of the VCF-formatted output (i.e. without file extension)
        output_format : str
            The format of the output (e.g. vcf, bcf, vcf.gz)
        keep_original : bool, optional 
            Decides if the original files should be kept
        optional_args : list, optional
            List of additional and optional arguments

    '''

    # Holds the arguments to convert to VCF format
    merge_args = ['merge']

    # Assign the output format arguments
    output_format_args = return_output_format_args(output_format)

    # Store the output format arguments
    merge_args.extend(output_format_args)

    # Stores the specified output filename
    vcf_output = '%s.%s' % (output_prefix, output_format)

    # Assigns the output file to the arguments
    merge_args.extend(['-o', vcf_output])

    # Loop the file names
    for filename in filenames:

        # Check if the files are indexed
        if check_for_index(filename) == False:

            # Create an index if not found
            create_index(filename)

    # Assigns the input files to merge
    merge_args.extend(filenames)

    # Call bcftools
    call_bcftools(merge_args + optional_args)

    # Delete the original files once the merged file is created
    if not keep_original:

        # Loop the files
        for filename in filenames:

            # Check if the file has an index
            if check_for_index(filename) == True:

                # Delete the index
                delete_index(filename)

            # Delete the file
            os.remove(filename)

def convert_vcf (filename, out_prefix, out_format, overwrite = False, keep_original = False):
    '''
        Converts a VCF-formatted file to VCF

        This function will convert a VCF-formatted file to VCF with the
        specified filename prefix. The function also has the option to keep or
        delete the input file once the VCF file has been created.

        Parameters
        ----------
        filename : str
            Filename of VCF-formatted input
        out_prefix : str
            Prefix of the VCF output (i.e. without file extension)
        out_format: str
            VCF-based file format for the output
        keep_original : bool, optional
            If the input file should be kept once converted
    '''

    logging.info('Beginning VCF conversion')

    # Check if the output format is VCF
    if out_format == 'vcf':

        # Stores the specified output_prefix to the VCF file
        vcf_output = '%s.vcf' % out_prefix

        # Holds the arguments to convert to VCF format
        convert_args = ['view', '-O', 'v']

    # Check if the output format is VCFGZ
    elif out_format == 'vcf.gz':

        # Stores the specified output_prefix to the VCFGZ file
        vcf_output = '%s.vcf.gz' % out_prefix

        # Holds the arguments to convert to VCFGZ format
        convert_args = ['view', '-O', 'z']

    # Check if the output format is BCF
    elif out_format == 'bcf':

        # Holds the arguments to convert to BCF format
        convert_args = ['convert', '-O', 'b']

        # Stores the specified output_prefix to the BCF file
        vcf_output = '%s.bcf' % out_prefix

    # Check if the previous output should not be overwritten
    if not overwrite:

        # Check if the file exists
        if os.path.isfile(vcf_output):
            raise Exception('%s already exists. Add --overwrite to ignore' % vcf_output)

    # Assigns the output file to the arguments
    convert_args.extend(['-o', vcf_output])

    # Assigns the specified input to the arguments
    convert_args.append(filename)

    # Call bcftools
    call_bcftools(convert_args)

    # Delete the original file once the vcf file is created
    if not keep_original:

        # Check if the file has an index
        if check_for_index(filename) == True:

            # Delete the index
            delete_index(filename)

        # Delete the file
        os.remove(filename)

    logging.info('Finished conversion')

def pipe_bcftools (bcftools_call_args):
    '''
        Calls bcftools with pipe output

        The output of this function is the stdout and stderr of bcftools. This
        function should only be used if bcftools is being used as the stdin of
        another function. Please note that this function does not check the for
        errors in the bcftools call. Please check for errors after the call is
        closed using check_bcftools_for_errors.

        Parameters
        ----------
        bcftools_stderr : str
            bcftools stderr

        Returns
        -------
        bcftools_call : PIPE
            Pipe of subprocess call, including both stdout and stderr

    '''

    # bcftools subprocess call
    bcftools_call = subprocess.Popen(['bcftools'] + list(map(str, bcftools_call_args)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    return bcftools_call

def pipe_bcftools_bcftools (bcftools_first_call_args, bcftools_second_call_args):
    '''
        Pipes the output of bcftools to bcftools

        The purpose of this function is to call bcftools multiple times. This
        is required to combine different bcftools functions or incompatible 
        arguments. Should consider updating the function to allow 2+ calls to
        bcftools.

        Parameters
        ----------
        bcftools_first_call_args : list
            first set of bcftools arguments
        bcftools_second_call_args : list
            second set of bcftools arguments

    '''
    
    # Open bcftools pipe
    bcftools_first_call = pipe_bcftools(bcftools_first_call_args)

    # bgzip subprocess call
    bcftools_second_call = subprocess.Popen(['bcftools'] + list(map(str, bcftools_second_call_args)), stdin = bcftools_first_call.stdout, stdout = subprocess.PIPE, stderr = subprocess.PIPE)

    # Wait for vctools to finish
    bcftools_first_call.wait()

    # Close the vcftools stdout
    bcftools_first_call.stdout.close()

    # Read the vcftools stderr
    bcftools_first_stderr = bcftools_first_call.stderr.read()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        bcftools_first_stderr = bcftools_first_stderr.decode()

    # Check that the log file was created correctly
    check_bcftools_for_errors(bcftools_first_stderr)

    # Wait for bgzip to finish
    bcftools_second_call.wait()

    # Save the stderr from second bcftools call
    bcftools_second_stdout, bcftools_second_stderr = bcftools_second_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        bcftools_second_stderr = bcftools_second_stderr.decode()

    # Check that output file was compressed correctly
    check_bcftools_for_errors(bcftools_second_stderr)

    logging.info('bcftools calls complete')

def pipe_bcftools_to_list (bcftools_call_args):
    '''
        Pipes output of bcftools to a set

        The purpose of this function is to return a set of the output of
        various bcftools commands, this will result in the removal of duplicate
        entries.

        Parameters
        ----------
        bcftools_call_args : list
            bcftools arguments

        Returns
        -------
        set_to_return : set
            Set of unique elements from the bcftools output
    '''

    # Open bcftools pipe
    bcftools_call = pipe_bcftools(bcftools_call_args)

    # Create a list to hold unique elements
    list_to_return = []

    try:

        # Iterate the bcftools stdout unless error occurs
        for bcftools_stdout_line in iter(bcftools_call.stdout.readline, b''):

            # Check if code is running in python 3
            if sys.version_info[0] == 3:
                # Convert bytes to string
                bcftools_stdout_line = bcftools_stdout_line.decode()

            # Remove the newline character
            bcftools_line = bcftools_stdout_line.strip()
            # Save the line to the list
            list_to_return.append(bcftools_line)

    except:
        raise Exception('bcftools call error')

    # Close the bcftools stdout
    bcftools_call.stdout.close()

    # Wait for bctools to finish
    bcftools_call.wait()

    # Read the bcftools stderr
    bcftools_stderr = bcftools_call.stderr.read()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        bcftools_stderr = bcftools_stderr.decode()

    # Check that the log file was created correctly
    check_bcftools_for_errors(bcftools_stderr)

    logging.info('bcftools call complete')

    return list_to_return

def pipe_bcftools_to_set (bcftools_call_args):
    '''
        Pipes output of bcftools to a set

        The purpose of this function is to return a set of the output of
        various bcftools commands, this will result in the removal of duplicate
        entries.

        Parameters
        ----------
        bcftools_call_args : list
            bcftools arguments

        Returns
        -------
        set_to_return : set
            Set of unique elements from the bcftools output
    '''

    # Open bcftools pipe
    bcftools_call = pipe_bcftools(bcftools_call_args)

    # Create a set to hold unique elements
    set_to_return = set()

    try:

        # Current element, reduces duplicates if VCF is sorted
        previous_element = None

        # Iterate the bcftools stdout unless error occurs
        for bcftools_stdout_line in iter(bcftools_call.stdout.readline, b''):

            # Check if code is running in python 3
            if sys.version_info[0] == 3:
                # Convert bytes to string
                bcftools_stdout_line = bcftools_stdout_line.decode()

            # Remove the newline character
            bcftools_line = bcftools_stdout_line.strip()
            # Check if the element is different from the previous element
            if bcftools_line != previous_element:
                # Store the new element for comparisons to reduce duplicates
                previous_element = bcftools_line
                # Save the element in the set
                set_to_return.add(bcftools_line)

    except:
        raise Exception('bcftools call error')

    # Close the bcftools stdout
    bcftools_call.stdout.close()

    # Wait for bctools to finish
    bcftools_call.wait()

    # Read the bcftools stderr
    bcftools_stderr = bcftools_call.stderr.read()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        bcftools_stderr = bcftools_stderr.decode()

    # Check that the log file was created correctly
    check_bcftools_for_errors(bcftools_stderr)

    logging.info('bcftools call complete')

    return set_to_return

def call_bcftools (bcftools_call_args):
    '''
        Calls bcftools

        The function calls bcftools.

        Parameters
        ----------
        bcftools_call_args : list
            bcftools arguments

        Returns
        -------
        vcftools_err : str
            vcftools log output

        Raises
        ------
        Exception
            If bcftools stderr returns an error
    '''

    # bcftools subprocess call
    bcftools_call = subprocess.Popen(['bcftools'] + list(map(str, bcftools_call_args)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # Wait for bcftools to finish
    bcftools_stdout, bcftools_stderr = bcftools_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        bcftools_stderr = bcftools_stderr.decode()

    check_bcftools_for_errors(bcftools_stderr)

    logging.info('bcftools call complete')

'''
Deprecated Code
'''

def convert_to_bcf (filename, output_prefix, overwrite = False, keep_original = False):
    '''
        Converts a VCF-formatted file to BCF

        This function will convert a VCF-formatted file to BCF with the
        specified filename prefix. The function also has the option to keep or
        delete the input file once the BCF file has been created.

        Parameters
        ----------
        filename : str
            Filename of VCF-formatted input
        output_prefix : str
            Prefix of the BCF output (i.e. without file extension)
        keep_original : bool, optional
            If the input file should be kept once converted
    '''

    logging.warning('Deprecated Code - Please use convert_vcf instead')

    # Stores the specified output_prefix to the BCF file
    bcf_output = '%s.bcf' % output_prefix

    # Check if the previous output should not be overwritten
    if not overwrite:

        # Check if the file exists
        if os.path.isfile(bcf_output):
            raise Exception('%s already exists. Add --overwrite to ignore' % bcf_output)

    # Holds the arguments to convert to BCF format
    convert_args = ['convert', '-O', 'b']

    # Assigns the output file to the arguments
    convert_args.extend(['-o', bcf_output])

    # Assigns the specified input to the arguments
    convert_args.append(filename)

    # Call bcftools
    call_bcftools(convert_args)

    # Delete the original file once the bcf file is created
    if not keep_original:
        
        # Check if the file has an index
        if check_for_index(filename) == True:

            # Delete the index
            delete_index(filename)

        # Delete the file
        os.remove(filename)

def convert_to_vcf (filename, output_prefix, overwrite = False, keep_original = False):
    '''
        Converts a VCF-formatted file to VCF

        This function will convert a VCF-formatted file to VCF with the
        specified filename prefix. The function also has the option to keep or
        delete the input file once the VCF file has been created.

        Parameters
        ----------
        filename : str
            Filename of VCF-formatted input
        output_prefix : str
            Prefix of the VCF output (i.e. without file extension)
        keep_original : bool, optional
            If the input file should be kept once converted
    '''

    logging.warning('Deprecated Code - Please use convert_vcf instead')

    # Stores the specified output_prefix to the VCF file
    vcf_output = '%s.vcf' % output_prefix

    # Check if the previous output should not be overwritten
    if not overwrite:

        # Check if the file exists
        if os.path.isfile(vcf_output):
            raise Exception('%s already exists. Add --overwrite to ignore' % vcf_output)

    # Holds the arguments to convert to VCF format
    convert_args = ['view', '-O', 'v']

    # Assigns the output file to the arguments
    convert_args.extend(['-o', vcf_output])

    # Assigns the specified input to the arguments
    convert_args.append(filename)

    # Call bcftools
    call_bcftools(convert_args)

    # Delete the original file once the vcf file is created
    if not keep_original:

        # Check if the file has an index
        if check_for_index(filename) == True:

            # Delete the index
            delete_index(filename)

        # Delete the file
        os.remove(filename)

def convert_to_vcfgz (filename, output_prefix, overwrite = False, keep_original = False):
    '''
        Converts a VCF-formatted file to bgzipped-VCF

        This function will convert a VCF-formatted file to bgzipped-VCF with the
        specified filename prefix. The function also has the option to keep or
        delete the input file once the bgzipped-VCF file has been created.

        Parameters
        ----------
        filename : str
            Filename of VCF-formatted input
        output_prefix : str
            Prefix of the bgzipped-VCF output (i.e. without file extension)
        keep_original : bool, optional
            If the input file should be kept once converted
    '''

    logging.warning('Deprecated Code - Please use convert_vcf instead')

    # Stores the specified output_prefix to the VCFGZ file
    vcfgz_output = '%s.vcf.gz' % output_prefix

    # Check if the previous output should not be overwritten
    if not overwrite:

        # Check if the file exists
        if os.path.isfile(vcfgz_output):
            raise Exception('%s already exists. Add --overwrite to ignore' % vcfgz_output)

    # Holds the arguments to convert to VCFGZ format
    convert_args = ['view', '-O', 'z']

    # Assigns the output file to the arguments
    convert_args.extend(['-o', vcfgz_output])

    # Assigns the specified input to the arguments
    convert_args.append(filename)

    # Call bcftools
    call_bcftools(convert_args)

    # Delete the original file once the vcfgz file is created
    if not keep_original:
        
        # Check if the file has an index
        if check_for_index(filename) == True:

            # Delete the index
            delete_index(filename)

        # Delete the file
        os.remove(filename)




#from .vcf_ref_to_seq import *
#from .vcf_from_regions import *
#from .vcf_reader_func import *
#from .gene_region import *
#from .logging_module import *
#from .four_gamete_pysam import *
#from .vcf_to_ima import *
#from .find_intergenic_bed import *
#from .get_nonmissing_chunks import *
#from .vcf_phase import *
#from .vcf_calc import *
#from .vcf_filter import *
#from .vcf_split import *
#from .vcftools import *
#from .bcftools import *

# Basic Information
__title__     = "py-popgen"
__version__   = '0.1.2'
__summary__   = "Software platform for facilitating population genomic analyses"
__url__       = "https://ppp.csusm.edu/"
__license__   = "MIT"
__copyright__ = "Copyright 2019, Arun Sethuraman and Jody Hey"

# Author Information
__author__    = "Andrew Webb, Jared Knoblauch, Nitesh Sabankar, " \
                "Apeksha Sukesh Kallur, Jody Hey and Arun Sethuraman"               
__email__     = "asethuraman@csusm.edu"







#!/usr/bin/env python
'''
    Automates the estimation of individual ancestries using Admixture. The
    functions allows for input as: i) Binary-PED files or ii) PED 12-formatted
    files. The function is also capable of configuring the optional arguments
    of Admixture.

    ##################
    Command-line Usage
    ##################
    The admixture automater may be called using the following command:

    .. code-block:: bash
        
        python admixture.py

    *************
    Example usage
    *************
    Estimating individual ancestries for each sample within *hapmap3.bed*
    for three ancestral populations.

    .. code-block:: bash
        
        python admixture.py --binary-ped-prefix hapmap3 --pop 3

    ############
    Dependencies 
    ############
    * `Admixture <http://software.genetics.ucla.edu/admixture/>`_

    ############################
    Input Command-line Arguments
    ############################
    **--ped-12-prefix** *<input_prefix>*
        Argument used to define the filename prefix shared by the 12-formatted ped 
        file (.ped) and the map file (.map). Should not be used alongside the 
        specific file arguments (e.g. --ped).
    **--ped-12** *<ped_filename>*
        Argument used to define the filename of the plink 12-formatted ped file 
        (.ped). Must be called alongside --map. Cannot be called alongside 
        --ped-prefix.
    **--map** *<map_filename>*
        Argument used to define the filename of the plink map file (.map). Must be 
        called alongside --ped. Cannot be called alongside --ped-prefix.
    **--binary-ped-prefix** *<input_prefix>*
        Argument used to define the filename prefix shared by the binary ped file
        (.bed), the fam file (.fam), and the bim file (.bim). Should not be used 
        alongside the specific file arguments (e.g. --binary-ped).
    **--binary-ped** *<binary_ped_filename>*
        Argument used to define the filename of the plink binary ped file (.bed). 
        Must be called alongside --fam and --bim. Cannot be called alongside 
        --binary-ped-prefix.
    **--fam** *<fam_filename>*
        Argument used to define the filename of the plink fam file (.fam). Must be 
        called alongside --binary-ped and --bim. Cannot be called alongside 
        --binary-ped-prefix.
    **--bim** *<bim_filename>*
        Argument used to define the filename of the plink bim file (.bim). Must be 
        called alongside --binary-ped and --fam. Cannot be called alongside 
        --binary-ped-prefix.

    #############################
    Output Command-line Arguments
    #############################
    **--overwrite**
        Argument used to define if previous output should be overwritten.

    ###############################
    Required Command-line Arguments
    ###############################
    **--pop** *<K_int>*
        Argument used to defines the number of ancestral populations.
    **--admix-method** *<em, block>*
        Argument used to define the algorithm to use. Two algorithm are supported: 
        Block relaxation algorithm (block) or EM algorithm (em). By default, the 
        Block relaxation algorithm is used.

    ###############################
    Optional Command-line Arguments
    ###############################
    **--acceleration** *<acceleration_int>*
        Argument used to defines the value of quasi-Newton acceleration method.
    **--major-converge-likelihood** *<likelihood_float>*
        Argument used to define the major terminaton criterion. Halt when the
        log-likelihood increases by less than the specified value between
        iterations.
    **--major-converge-iter** *<iter_int>*
        Argument used to define the major terminaton criterion. Defines the
        maximum number of iterations.
    **--minor-converge-likelihood** *<likelihood_float>*
        Argument used to define the minor terminaton criterion. Halt when the
        log-likelihood increases by less than the specified value between
        iterations.
    **--minor-converge-iter** *<iter_int>*
        Argument used to define the minor terminaton criterion. Defines the
        maximum number of iterations.
    **--bootstrap** *<bootstrap_int>*
        Argument used to define the number of bootstrap replicates.
    **--random-seed** *<seed_int>*
        Argument used to define the seed value for the random number generator.
    **--threads** *<thread_int>*
        Argument used to define the number of threads to be used for computation.
'''

import os
import argparse
import logging
import subprocess

from pgpipe.logging_module import initLogger, logArgs
from pgpipe.plink import confirm_ped_prefix, confirm_bed_prefix, confirm_ped_files, confirm_bed_files

def admix_parser(passed_arguments):
    '''admix Argument Parser - Assigns arguments from command line'''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    admix_parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # Input PED/BED prefix arguments
    admix_prefix = admix_parser.add_mutually_exclusive_group()
    admix_prefix.add_argument("--ped-12-prefix", dest = 'ped_prefix', help = "Defines the filename prefix of both PED-12 and MAP files", type = str)
    admix_prefix.add_argument("--binary-ped-prefix", dest = 'bed_prefix', help = "Defines the filename prefix of the Binary-PED, FAM, and BIM files", type = str)

    # Input PED arguments
    admix_parser.add_argument("--ped-12", dest = 'ped_filename', help = "Defines the filename of the PED-12 file. Called alongside --map", type = str, action = parser_confirm_file())
    admix_parser.add_argument("--map", dest = 'map_filename', help = "Defines the filename of the MAP file. Called alongside --ped-12", type = str, action = parser_confirm_file())

    # Input BED arguments
    admix_parser.add_argument("--binary-ped", dest = 'bed_filename', help = "Defines the filename of the Binary-PED (i.e. BED) file. Called alongside --fam and --bim", type = str, action = parser_confirm_file())
    admix_parser.add_argument("--fam", dest = 'fam_filename', help = "Defines the filename of the FAM file. Called alongside --binary-ped and --bim", type = str, action = parser_confirm_file())
    admix_parser.add_argument("--bim", dest = 'bim_filename', help = "Defines the filename of the BIM file. Called alongside --binary-ped and --fam", type = str, action = parser_confirm_file())

    # Population argument
    admix_parser.add_argument('--pop', help = 'Defines the number of populations', required = True, type = int)

    # Method argument
    method_list = ['em', 'block']
    method_default = 'block'
    admix_parser.add_argument('--admix-method', metavar = metavar_list(method_list), help = 'Defines the algorithm to use. Default: block', type = str, choices = method_list, default = method_default)

    # Acceleration arguments
    admix_parser.add_argument('--acceleration', help = 'Defines the quasi-Newton acceleration value', type = int)

    # Convergence arguments
    admix_major = admix_parser.add_mutually_exclusive_group()
    admix_major.add_argument('--major-converge-likelihood', dest = 'major_converge', help = 'Defines the major terminaton criterion for the log-likelihood', type = float)
    admix_major.add_argument('--major-converge-iter', dest = 'major_converge', help = 'Defines the major terminaton criterion for the maximum number of iterations.', type = int)
    admix_minor = admix_parser.add_mutually_exclusive_group()
    admix_minor.add_argument('--minor-converge-likelihood', dest = 'minor_converge', help = 'Defines the minor convergence criterion for the log-likelihood', type = float)
    admix_minor.add_argument('--minor-converge-iter', dest = 'minor_converge', help = 'Defines the minor convergence criterion for the maximum number of iterations', type = int)

    # General arguments
    admix_parser.add_argument('--bootstrap', help = 'Defines the number of bootstrap replicates', type = int)
    admix_parser.add_argument('--threads', help = 'Defines the number of threads to be used for computation', type = int)
    admix_parser.add_argument('--random-seed', help = "Defines the seed value for the random number generator", type = int)
    admix_parser.add_argument('--overwrite', help = "Defines if previous output should be overwritten", action = 'store_true')


    if passed_arguments:
        return admix_parser.parse_args(passed_arguments)
    else:
        return admix_parser.parse_args()


def run(passed_arguments = []):

	# Assign location of admixture file
    admixture_exec = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),'bin','admixture')

    # Check that admixture exists at specified path
    if not os.path.isfile(admixture_exec):
        raise IOError('admixture executable not found in Path specified: %s' % admixture_exec)


    # Grab admixture arguments from command line
    admix_args = admix_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(admix_args, func_name = 'admixture')    

    # List to store arguments to be passed to admixture
    admix_call_args = []

    # Check if a method not has been assigned. Error should not be seen
    if not admix_args.admix_method:
        raise Exception('Algorithm assignment error. Please contact development team')

    # Assign the algorithm to the argument list
    admix_call_args.append('--method=' + admix_args.admix_method)

    # Check if the user has assigned an acceleration value
    if admix_args.acceleration:

        # Assign the acceleration value to the argument list
        admix_call_args.extend(['-a', 'qn' + str(admix_args.acceleration)])

    # Check if the user assigned a major terminaton criterion
    if admix_args.major_converge:

        # Assign the criterion to the argument list
        admix_call_args.append('-C=' + str(admix_args.major_converge))

    # Check if the user assigned a minor terminaton criterion
    if admix_args.minor_converge:

        # Assign the criterion to the argument list
        admix_call_args.append('-c=' + str(admix_args.minor_converge))

    # Check if a random seed has been assigned
    if admix_args.random_seed:

        # Assign the seed value to the argument list
        admix_call_args.append('--seed=' + str(admix_args.random_seed))

    # Check if the bootstrap replicates have been assigned 
    if admix_args.bootstrap:

        # Assign the number of bootstrap replicates to the argument list
        admix_call_args.append('-B' + str(admix_args.bootstrap))
    
    # Check if the user has defined the number of threads to use
    if admix_args.threads:

        # Assign the number of threads to the argument list
        admix_call_args.append('-j' + str(admix_args.threads))

    # Check if the a ped prefix was assigned
    if admix_args.ped_prefix and confirm_ped_prefix(admix_args.ped_prefix):

        # Add the ped-12 filename to the argument list
        admix_call_args.append(admix_args.ped_prefix + '.ped')

    # Check if the a bed prefix was assigned
    elif admix_args.bed_prefix and confirm_bed_prefix(admix_args.bed_prefix):

        # Add the binary-ped filename to the argument list
        admix_call_args.append(admix_args.bed_prefix + '.bed')

    # Check if ped files were assigned
    elif confirm_ped_files (admix_args.ped_filename, admix_args.map_filename):

    	# Check if the files have the same prefix. Update later
    	if admix_args.ped_filename[:-4] != admix_args.map_filename[:-4]:
    		raise Exception('PED and MAP files must share the same prefix - i.e. input.ped/input.map')

    	# Add the ped-12 filename to the argument list
    	admix_call_args.append(admix_args.ped_filename)

    # Check if bed files were assigned
    elif confirm_bed_files (admix_args.bed_filename, admix_args.bim_filename, admix_args.fam_filename):

    	# Check if the files have the same prefix. Update later
    	if admix_args.bed_filename[:-4] != admix_args.bim_filename[:-4] or admix_args.bed_filename[:-4] != admix_args.fam_filename[:-4]:
    		raise Exception('Binary-PED, BIM, and FAM files must share the same prefix - i.e. input.bed/input.bim/input.fam')

    	# Add the ped-12 filename to the argument list
    	admix_call_args.append(admix_args.bed_filename)

    else:

    	raise Exception('No input specified. Please check command-line')

    # Add the number of populations to the argument list
    admix_call_args.append(admix_args.pop)

    logging.info('admixture parameters assigned')

    # Run 'admixture' executable file with options provided by user
    admixture_call = subprocess.Popen([admixture_exec] + list(map(str, admix_call_args)), stdout = subprocess.PIPE, stderr = subprocess.PIPE)

    # Store command output and/or error to variables
    admix_stdout, admix_stderr = admixture_call.communicate()

    # Check if admixture returned an error
    if admix_stderr:
        raise Exception(admix_stderr)

if __name__ == "__main__":
    initLogger()
    run()



import os
import sys
import subprocess
import logging
import fileinput

def log_bedtools_reference (out_filename, append_mode = False, ref_header = True):

    # Check if the file is to be written in append mode
    if append_mode:

        # Open the file
        log_file = open(out_filename + '.log', 'a')

        # Check if the ref header should be added
        if ref_header:
            log_file.write('\nPlease Reference alongside the PPP:\n')

    else:

        # Open the file
        log_file = open(out_filename + '.log', 'w')

        # Check if the ref header should be added
        if ref_header:
            log_file.write('Please Reference alongside the PPP:\n')

    log_file.write('')
    log_file.close()

    logging.info('Reference assigned')

def check_bedtools_for_errors (bedtools_stderr):
    '''
        Checks the bedtools stderr for errors

        Parameters
        ----------
        bedtools_stderr : str
            bedtools stderr

        Raises
        ------
        IOError
            If bedtools stderr returns an error
    '''

    # Print output if error found. Build up as errors are discovered
    if bedtools_stderr:

        # Splits log into list of lines
        bedtools_stderr_lines = bedtools_stderr.splitlines()

        # Prints the error(s)
        raise Exception('\n'.join((output_line.split(':')[1] for output_line in bedtools_stderr_lines if 'ERROR:' in output_line)))

def merge_bed_files (bed_files, bed_output_filename, optional_merge_args):
	'''
		Merge BED files

		The function calls bedtools

		Parameters
		----------
		bedtools_call_args : list
			bedtools arguments

		Raises
		------
		Exception
			If bedtools stderr returns an error
	'''

	# Assign the BEDtools input using fileinput
	bed_input = fileinput.input(files = bed_files, mode = 'rb')

	# Create the output file
	bed_output_file = open(bed_output_filename, 'w')

	# Call BEDtools to sort the input files
	bedtools_sort_call = subprocess.Popen(['bedtools', 'sort', '-i', '-'], stdin = subprocess.PIPE, stdout = subprocess.PIPE, stderr = subprocess.PIPE)

	# Loop the BEDtools input
	for bed_line in bed_input:

		# Pipe the BEDtools input to stdin
		bedtools_sort_call.stdin.write(bed_line)

	# Call BEDtools to merge the sorted input
	bedtools_merge_call = subprocess.Popen(['bedtools', 'merge', '-i', '-'] + list(map(str, optional_merge_args)), stdin = bedtools_sort_call.stdout, stdout = bed_output_file, stderr = subprocess.PIPE)

	# Close the sort stdin
	bedtools_sort_call.stdin.close()

	# Close the sort stdout
	bedtools_sort_call.stdout.close()

	# Read the sort stderr
	bedtools_sort_stderr = bedtools_sort_call.stderr.read()

	# Check if code is running in python 3
	if sys.version_info[0] == 3:

		# Convert bytes to string
		bedtools_sort_stderr = bedtools_sort_stderr.decode()

	# Check BEDtools sort for possible errors
	check_bedtools_for_errors(bedtools_sort_stderr)

	# Wait for sort call to finish
	bedtools_sort_call.wait()

	# Save the stderr from bgzip, stdout = None
	bedtools_merge_stdout, bedtools_merge_stderr = bedtools_merge_call.communicate()

	# Check if code is running in python 3
	if sys.version_info[0] == 3:

		# Convert bytes to string
		bedtools_merge_stderr = bedtools_merge_stderr.decode()

	# Check BEDtools sort for possible errors
	check_bedtools_for_errors(bedtools_merge_stderr)

	# Close BEDtools output
	bed_output_file.close()


def standard_bedtools_call (bedtools_call_args, bed_output_filename):
    '''
        Standard call of bedtools

        The function calls bedtools. Returns the stderr of bedtools to create a log
        file of the call.

        Parameters
        ----------
        bedtools_call_args : list
            bedtools arguments

        Raises
        ------
        Exception
            If bedtools stderr returns an error
    '''

    # Create a string with the bedtools path

    #bedtools_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'bin', 'bedtools')

    # Create the output file
    bed_output_file = open(bed_output_filename, 'w')

    # bedtools subprocess call
    bedtools_call = subprocess.Popen(['bedtools'] + list(map(str, bedtools_call_args)), stdout = bed_output_file, stderr = subprocess.PIPE)

    # Wait for bedtools to finish
    bedtools_out, bedtools_err = bedtools_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        bedtools_err = bedtools_err.decode()

    logging.info('bedtools call complete')

    # Check that the log file was created correctly
    check_bedtools_for_errors(bedtools_err)


#!/usr/bin/env python
'''
    Automates various utilites for VCF-formatted files. This currently includes:
    obtain a list of the chromosomes within a VCF-based file, obtain a list of the 
    samples within a VCF-based file, concatenate multiple VCF-based files, merge
    multiple VCF-based files, and sort a VCF-based file.

    ##################
    Command-line Usage
    ##################
    The VCF utilites function may be called using the following command:

    .. code-block:: bash
        
        python vcf_utilites.py

    *************
    Example usage
    *************
    Concatenate multiple VCF files:

    .. code-block:: bash
        
        python vcf_utilites.py --vcfs chr21.vcf.gz chr22.vcf.gz --utility concatenate

    Merge multiple VCF files:

    .. code-block:: bash
        
        python vcf_utilites.py --vcfs chr22.ceu.vcf.gz chr22.yri.vcf.gz --utility merge
    
    ############
    Dependencies 
    ############
    * `BCFtools <https://samtools.github.io/bcftools/bcftools.html>`_
    
    ############################
    Input Command-line Arguments
    ############################
    **--vcf** *<input_filename>*
        Argument used to define the filename of the VCF file.
    **--vcfs** *<input_filename>* *<input1_filename, input2_filename, etc.>*
        Argument used to define the filename of the VCF file(s). May be used multiple 
        times.
    
    #############################
    Output Command-line Arguments
    #############################
    **--out** *<output_filename>*
        Argument used to define the complete output filename, overrides **--out-prefix**.
    **--out-prefix** *<output_prefix>*
        Argument used to define the output prefix (i.e. filename without file extension)
    **--overwrite**
        Argument used to define if previous output should be overwritten.
    
    ##################################
    Utility Command-line Specification
    ##################################
    **--utility** *<sample-list, chr-list, concatenate, merge, sort>*
        Argument used to define the desired utility. Current utilities include: creation
        of a file of the samples within the VCF (sample-list); creation of a file of the 
        chromosomes within the VCF (chr-list); combine multiple VCF files with different
        variants but the same samples (concatenate); combine multiple VCF files with 
        different samples but the same variants (merge); or sort a single VCF file (sort).

    *****************************************
    Additional Utility Command-line Arguments
    *****************************************
    **--record-merge-mode** *<none, snps, indels, both, all, id>*
        Argument used to define the type of multiallelic records to create. Only usable with 
        the merge utility.
    **--record-missing-as-ref**
        Argument used to define that missing records should be converted to the reference
        allele. Only usable with the merge and concatenate utilites.
    **--out-format** *<vcf, vcf.gz, bcf>*
        Argument used to define the desired output format. Formats include: uncompressed 
        VCF (vcf); compressed VCF (vcf.gz) [default]; and BCF (bcf). Only usable with the 
        merge and concatenate utilites.
'''

import os
import sys
import argparse
import itertools
import copy
import shutil
import logging
import pandas as pd

# Import basic vcftools functions
from pgpipe.bcftools import *

# Insert Jared's directory path, required for calling Jared's functions. Change when directory structure changes.
#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir, 'pppipe')))

from pgpipe.logging_module import initLogger, logArgs

def vcf_utility_parser(passed_arguments):
    '''
    VCF Utility Argument Parser

    Assign the parameters for VCF Utility using argparse.

    Parameters
    ----------
    passed_arguments : list, optional
        Parameters passed by another function. sys.argv is used if
        not given. 

    Raises
    ------
    IOError
        If the input, or other specified files do not exist
    '''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def parser_confirm_file_list ():
        '''Custom action to confirm file exists in list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                # Loop the list
                for value_item in value:
                    # Check if the file exists
                    if not os.path.isfile(value_item):
                        raise IOError('%s not found' % value_item)
                if not getattr(args, self.dest):
                    setattr(args, self.dest, value)
                else:
                    getattr(args, self.dest).extend(value)
        return customAction

    def parser_add_to_list ():
        '''Custom action to add items to a list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not getattr(args, self.dest):
                    setattr(args, self.dest, value)
                else:
                    getattr(args, self.dest).extend(value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    vcf_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # Input arguments.
    vcf_input = vcf_parser.add_mutually_exclusive_group(required=True)
    vcf_input.add_argument('--vcf', help = "Defines the filename of the VCF", type = str, action = parser_confirm_file())
    vcf_input.add_argument('--vcfs', help = "Defines the filenames of the VCFs (may be used multiple times)", type = str, nargs = '+', action = parser_confirm_file_list())

    # Output file arguments
    vcf_parser.add_argument('--out', help = 'Defines the complete output filename, overrides --out-prefix', type = str)
    vcf_parser.add_argument('--out-prefix', help = 'Defines the output prefix (i.e. filename without file extension)', default = 'out')
    vcf_parser.add_argument('--overwrite', help = "Defines that previous output files should be overwritten", action = 'store_true')
    
    # Utility based arguments
    utility_list = ['sample-list', 'chr-list', 'concatenate', 'merge', 'sort']
    vcf_parser.add_argument('--utility', metavar = metavar_list(utility_list), help = 'Defines the utility to use', type = str, choices = utility_list)
    
    merge_list = ['none', 'snps', 'indels', 'both', 'all', 'id']
    vcf_parser.add_argument('--record-merge-mode', metavar = metavar_list(merge_list), help = 'Defines the type of multiallelic records to create. Only usable with the merge utility', type = str, choices = merge_list)
    
    vcf_parser.add_argument('--record-missing-as-ref', help = 'Defines that missing records should be converted to the reference allele. Only usable with the merge and concatenate utilites', action = 'store_true')
    
    out_format_list = ['vcf', 'vcf.gz', 'bcf']
    out_format_default = 'vcf.gz'
    vcf_parser.add_argument('--out-format', metavar = metavar_list(out_format_list), help = 'Defines the desired output format', type = str, choices = out_format_list, default = out_format_default)
    
    if passed_arguments:
        return vcf_parser.parse_args(passed_arguments)
    else:
        return vcf_parser.parse_args()

def utility_tsv_output (utility_results, utility_header, utility_index, utility_output_filename):

    # Check if the utility didn't produced results
    if not utility_results:
        raise Exception('Utility (%s) produced no results' % vcf_args.utility)

    # Check if index values were assigned, and if so, create file with index
    if utility_index:

        # Save the utility information into a dataframe
        utility_dataframe = pd.DataFrame(utility_results, index = utility_index, columns = utility_header)

        # Save the utility dataframe into a file
        utility_dataframe.to_csv(utility_output_filename, sep = '\t')

    # Create file without index
    else:

        # Save the utility information into a dataframe
        utility_dataframe = pd.DataFrame(utility_results, columns = utility_header)

        # Save the utility dataframe into a file
        utility_dataframe.to_csv(utility_output_filename, sep = '\t', index = False)

def run (passed_arguments = []):
    '''
    Utilites for VCF-formatted files

    This function uses the argparse-based function :py:func:`vcf_utility_parser`
    to parse either sys.argv or passed_arguments to obtain the parameters below. 
    The parameters are then translated to their BCFtools equivalent. Once all the 
    parameters are assigned, BCFtools is called.

    Parameters
    ----------
    --vcf : str
        Filename of the VCF
    --vcfs : list
        List of VCF filenames
    --utility : str
        Utility to be used
    --out : str
        Complete output filename, overrides --out-prefix
    --out-prefix : str
        Output prefix
    --out-format : str
        Desired output format
    --overwrite
        Species if previous output should be overwriten
    --record-merge-mode : str
        Type of multiallelic records to create
    --record-missing-as-ref : bool
        Convert missing records to the reference allele

    Raises
    ------
    IOError
        Output file already exists
    '''

    # Grab VCF arguments from command line
    vcf_args = vcf_utility_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(vcf_args, func_name = 'vcf_utilities')

    # Check if the function does not support multiple vcf files
    if vcf_args.utility in ['chr-list', 'sample-list', 'sort']:

        # Check if multiple vcf files have been assigned
        if vcf_args.vcfs:
            raise Exception('Utility (%s) only supports a single VCF file' % vcf_args.utility)

    # Check if the function does not support a single vcf file
    if vcf_args.utility in ['merge', 'concatenate']:

        # Check if multiple vcf files have been assigned
        if vcf_args.vcf:
            raise Exception('Utility (%s) requires multiple VCF files' % vcf_args.utility)

    # Assign the standard output filename
    utility_output_filename = vcf_args.out_prefix

    # Check if the user has requested a chromosome list
    if vcf_args.utility == 'chr-list':

        # Assign the suffix
        utility_output_filename += '.chr.list'

    # Check if the user has requested a sample list
    elif vcf_args.utility == 'sample-list':

        # Assign the suffix
        utility_output_filename += '.sample.list'

    # Check if the user has requested to merge or concatenate VCFs
    elif vcf_args.utility in ['merge', 'concatenate', 'sort']:

        # Assign the suffix
        utility_output_filename += '.%s' % vcf_args.out_format

    # Check if the user specified a output filename
    if vcf_args.out:

        # Override the out_prefix output filename
        utility_output_filename = vcf_args.out

    # Check if previous output should not be overwriten
    if not vcf_args.overwrite:

        # Check if previous output exists
        if os.path.exists(utility_output_filename):
            raise IOError('Utility output already exists')

    # List of hold optional utility-based arguments
    utility_optional_args = []

    if vcf_args.utility in ['merge', 'concatenate']:

        # Check if the user has requested to convert missing data into reference alleles
        if vcf_args.record_missing_as_ref:

            # Assign the optional argument
            utility_optional_args.append('--missing-to-ref')

    if vcf_args.utility == 'merge':

        # Check if the user has selected a merge mode
        if vcf_args.record_merge_mode:

            # Assign the optional argument
            utility_optional_args.extend(['--merge', vcf_args.record_merge_mode])

    # List of hold utility results
    utility_results = []

    # List of the utility index
    utility_index = []

    # Str of the utility header
    utility_header = ''

    # Check if the user has requested a chromosome list
    if vcf_args.utility == 'chr-list':

        # Use bcftools to assign the unique chromosomes within the vcf
        utility_results = get_unique_chrs(vcf_args.vcf)

        # Assign the chromosome list header
        utility_header = ['Chromosomes']

        # Create an output file for the utility
        utility_tsv_output(utility_results, utility_header, utility_index, utility_output_filename)

    # Check if the user has requested a sample list
    elif vcf_args.utility == 'sample-list':

        # Use bcftools to assign the samples within the vcf
        utility_results = get_samples(vcf_args.vcf)

        # Assign the chromosome list header
        utility_header = ['Samples']

        # Create an output file for the utility
        utility_tsv_output(utility_results, utility_header, utility_index, utility_output_filename)

    # Check if the user has requested a concatenated VCF
    elif vcf_args.utility == 'concatenate':

        # Concatenate the VCF files
        concatenate(vcf_args.vcfs, vcf_args.out_prefix, vcf_args.out_format, keep_original = True, optional_args = utility_optional_args)

    # Check if the user has requested a concatenated VCF
    elif vcf_args.utility == 'merge':

        # Concatenate the VCF files
        merge(vcf_args.vcfs, vcf_args.out_prefix, vcf_args.out_format, keep_original = True, optional_args = utility_optional_args)

    # Check if the user has requested a concatenated VCF
    elif vcf_args.utility == 'sort':

        # Assign the basic arguments for sort
        sort_arguments = ['sort', '-o', utility_output_filename]

        # Assign the output format arguments
        sort_arguments += return_output_format_args(vcf_args.out_format)

        # Append the input VCF
        sort_arguments.append(vcf_args.vcf)

        # Concatenate the VCF files
        call_bcftools(sort_argumentse)

    # Check if the output needs to renamed to a user specified filename
    if vcf_args.out and vcf_args.utility in ['merge', 'concatenate']:

        # Override the out_prefix output filename
        shutil.move('%s.%s' % (vcf_args.out_prefix, vcf_args.out_format), utility_output_filename)


if __name__ == "__main__":
    #initLogger()
    run()



import os
import sys
import subprocess
import shutil
import argparse
import logging
import copy

# Call PPP-based scripts
#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir,'pppipe')))

from pgpipe.logging_module import initLogger, logArgs
from pgpipe.fasta import check_format

def check_ref_file_association (ref_filename, dict_filename, dictonary_ext = 'dict', gunzip_ext = 'gz'):

    # Check the format of the reference file
    ref_format = check_format(ref_filename)

    # Get the basename of the reference
    ref_basename = os.path.basename(ref_filename)

    # Check if the file is fasta formatted
    if ref_format == 'fasta':

        # Check if the file has an extension
        if ref_basename.count(os.extsep) != 0:

            # Split the filename to seperate an ext
            ref_filename_wo_ext, ref_ext = ref_filename.split(os.extsep, 1)

            # Check if the dictonary file is correctly associated
            if dict_filename == ref_filename_wo_ext + os.extsep + dictonary_ext:
                return True

    # Check if the file is fasta formatted
    elif ref_format == 'gzip':

        # Check if the file has the gunzip file extension
        if ref_basename.endswith(gunzip_ext) and ref_basename.count(os.extsep) >= 2: 

            # Split the filename to seperate an ext
            ref_filename_wo_ext, ref_ext, gz_ext = ref_filename.split(os.extsep, 2)

            # Save the filename of the possible dictonary file
            dictonary_filename = ref_filename_wo_ext + os.extsep + dictonary_ext

            # Check if the dictonary file is correctly associated
            if dict_filename == ref_filename_wo_ext + os.extsep + dictonary_ext:
                return True

    # Return None, if association failed
    return None

def get_ref_dictonary (ref_filename, dictonary_ext = 'dict'):

    # Save the filename of the possible dictonary file
    dictonary_filename = ref_filename + os.extsep + dictonary_ext

    # Check if the dictonary file exists
    if os.path.isfile(dictonary_filename):

        # Return the filename
        return dictonary_filename

    # Save the filename before split
    previous_filename = copy.deepcopy(ref_filename)

    # Split the filename to seperate an ext
    ref_filename_wo_ext, ref_ext = ref_filename.split(os.extsep, 1)

    # Loop the reference filename until there are no more extentions
    while ref_filename_wo_ext != previous_filename:

        # Save the filename of the possible dictonary file
        dictonary_filename = ref_filename_wo_ext + os.extsep + dictonary_ext

        # Check if the dictonary file exists
        if os.path.isfile(dictonary_filename):

            # Return the filename
            return dictonary_filename

        # Split the file
        else:

            # Save the filename before split
            previous_filename = copy.deepcopy(ref_filename_wo_ext)

            # Split the filename to seperate an ext
            ref_filename_wo_ext, ref_ext = ref_filename.split(os.extsep, 1)

    # Return None, if no file is found
    return None

def produce_picard_log (output, filename):
    '''
        Creates the picard log file

        This function will create a log file from the picard stderr. Please
        run `check_picard_for_errors` prior to check that picard finished
        without error.

        Parameters
        ----------
        output : str
            picard stderr
        filename : str
            Specifies the filename for the log file

        Returns
        -------
        output : file
            picard log file

    '''

    picard_log_file = open(filename + '.log','w')
    picard_log_file.write(str(output))
    picard_log_file.close()

def remove_picard_index (out_filename, out_format):
    '''
        Removes the picard index file

        This function will remove the index file created by picard.

        Parameters
        ----------
        out_filename : str
            Specifies the filename for the vcf output file
        out_format : str 
            Specifies the file format of the vcf output

    '''
    
    # Check if the file is a gzipped VCF
    if out_format == 'vcf.gz':

        # Remove the index, if found
        if os.path.isfile(out_filename + '.tbi'):
            os.remove(out_filename + '.tbi')

    # Check if the file is a gzipped VCF
    if out_format == 'vcf':

        # Remove the index, if found
        if os.path.isfile(out_filename + '.idx'):
            os.remove(out_filename + '.idx')

    # Check if the file is a gzipped VCF
    if out_format == 'bcf':

        # Remove the index, if found
        if os.path.isfile(out_filename + '.idx'):
            os.remove(out_filename + '.idx')


def check_picard_for_errors (picard_stderr):
    '''
        Checks the picard stderr for errors

        Parameters
        ----------
        picard_stderr : str
            picard stderr

        Raises
        ------
        IOError
            If picard stderr returns an error
    '''

    # Check if there was no error
    if 'Elapsed time:' in picard_stderr:
        pass

    # Print exception output
    elif 'Exception' in picard_stderr:

        # Loop stderr lines
        for picard_stderr_line in picard_stderr.splitlines():

            # Check if the current line has the exception
            if 'Exception' in picard_stderr_line:
                raise Exception(picard_stderr_line)

    # Print error output. Build up as errors are discovered
    elif picard_stderr:
        raise Exception(picard_stderr)

def standard_picard_call (picard_path, picard_call_args, output_filename):
    '''
        Standard call of picard

        The function calls picard. Returns the stderr of picard to create a log
        file of the call.

        Parameters
        ----------
        picard_path : str
            Path to picard.jar
        picard_call_args : list
            picard arguments
        output_filename : str
            Output filename string

        Raises
        ------
        Exception
            If picard stderr returns an error
    '''

    # Assign location of picard jar file
    if picard_path is None:
        picard_jar = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),'bin','picard.jar')
    else:
        picard_jar = os.path.join(picard_path, 'picard.jar')

    # Check that picard.jar exists
    if not os.path.isfile(picard_jar):
        raise IOError('picard.jar not found. Path specified: %s' % picard_path)

    logging.info('picard parameters assigned')

    # Phasing subprocess call
    picard_call = subprocess.Popen(['java', '-jar', picard_jar] + picard_call_args, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    picard_stdout, picard_stderr = picard_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        picard_stderr = picard_stderr.decode()

    # Check picard call for errors
    check_picard_for_errors(picard_stderr)

    logging.info('picard call complete')

    # Create the log file
    produce_picard_log (picard_stderr, output_filename)

    logging.info('picard log created')


def call_picard (picard_path, picard_call_args, output_filename):
    '''
        Automates picard calls

        This function passes the argument list to standard_picard_call.

        Parameters
        ----------
        picard_path : str
            Path to picard.jar
        picard_call_args : list
            Argument list for picard
        output_filename : str
            Output filename string
    ''' 

    # Standard call to picard
    standard_picard_call(picard_path, picard_call_args, output_filename)







import os
import sys
import argparse
import logging
import subprocess
import random
import string
import shutil

import rpy2.robjects as robjects

from rpy2.robjects.packages import importr
from rpy2.robjects import pandas2ri

# Import PPP modules and scripts
from pgpipe.plink import *
from pgpipe.model import read_model_file, pops_not_in_model
from pgpipe.logging_module import initLogger, logArgs

def check_convertf_for_errors (convertf_stderr):
    '''
        Checks the convertf stderr for errors

        Parameters
        ----------
        convertf_stderr : str
            convertf stderr

        Raises
        ------
        IOError
            If convertf stderr returns an error
    '''
    
    # Check if IDs are too long
    if 'too long' in convertf_stderr:
        raise Exception('Individual and/or family names too long. Eigenstrat supports a maximum of 39 characters')

    # Expand as needed
    elif convertf_stderr:
        raise Exception(convertf_stderr)

def call_convertf (convertf_call_args):
    '''
        Calls convertf

        The function calls convertf. Returns the stderr of convertf to
        create log file of the call.

        Parameters
        ----------
        convertf_call_args : list
            convertf arguments

        Returns
        -------
        convertf_err : str
            convertf log output
    '''

    # convertf subprocess call without stdout
    convertf_call = subprocess.Popen(['convertf'] + list(map(str, convertf_call_args)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # Wait for convertf to finish
    convertf_stdout, convertf_stderr = convertf_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        convertf_stderr = convertf_stderr.decode()

    logging.info('convertf call complete')

    # Check that the log file was created correctly
    check_convertf_for_errors(convertf_stderr)

    return convertf_stderr

def assign_unique_filename (out_prefix, out_suffix, str_size = 10):

    # Generate a random string for par filename
    random_str = ''.join(random.choice(string.ascii_uppercase + string.digits) for digit in range(str_size))

    # Save the unique filename
    unique_filename = '%s.%s.%s' % (out_prefix, random_str, out_suffix)

    # Loop until filename is unique
    while os.path.isfile(unique_filename):

        # Generate a random string for par filename
        random_str = ''.join(random.choice(string.ascii_uppercase + string.digits) for digit in range(str_size))

        # Save the unique filename
        unique_filename = '%s.%s.%s' % (out_prefix, random_str, out_suffix)

    return unique_filename

def par_assign_input (gt_filename, snp_filename, ind_filename):
    
    # Assign the genotype filename
    par_input_str =  'genotypename:    %s\n' % gt_filename

    # Assign the snpname filename
    par_input_str += 'snpname:         %s\n' % snp_filename

    # Assign the indivname filename
    par_input_str += 'indivname:       %s\n' % ind_filename

    return par_input_str

def par_assign_eigenstrat (out_prefix):

    # Confirm an output prefix has been assigned
    if not out_prefix:
        raise Exception('No output prefix assigned. Unable to complete conversion')

    # Assign the output filenames
    par_eigenstrat_str =  'outputformat:    EIGENSTRAT\n'
    par_eigenstrat_str += 'genotypeoutname: %s.geno\n' % out_prefix
    par_eigenstrat_str += 'snpoutname:      %s.snp\n' % out_prefix
    par_eigenstrat_str += 'indivoutname:    %s.ind\n' % out_prefix

    return par_eigenstrat_str

def par_assign_family_ids (family_id):

    # Check if the user has requested family IDs
    if family_id:

        # Add the family ID option
        family_id_str =  'familynames:     YES'
        
    else:

        # Add the family ID option
        family_id_str = 'familynames:     NO'

    return family_id_str

def ped_to_eigenstrat (ped_filename = None, map_filename = None, ped_prefix = None, out_prefix = None, family_id = False, overwrite = False, **kwargs):

    # Check if ped-based files were assigned
    if ped_prefix and confirm_ped_prefix(ped_prefix):

        # Assign the files
        ped_filename = ped_prefix + '.ped'
        map_filename = ped_prefix + '.map'

    # Check that all the ped files are assigned
    elif ped_filename and confirm_ped_files(ped_filename, map_filename):

        # No action requried, just needs to pass
        pass 
   
    # Assign the par filename
    par_filename = assign_unique_filename(out_prefix, 'par')

    # Assign the input files
    par_file_str = par_assign_input(ped_filename, map_filename, ped_filename)

    # Add the Eigenstrat assignment lines
    par_file_str += par_assign_eigenstrat(out_prefix)

    # Add the family ID line
    par_file_str += par_assign_family_ids(family_id)

    # Create the par file
    par_file = open(par_filename, 'w')
    par_file.write(par_file_str)
    par_file.close()

    # Call convertf
    call_convertf(['-p', par_filename])

    # Remove the par file
    os.remove(par_filename)

def bed_to_eigenstrat (bed_filename = None, bim_filename = None, fam_filename = None, bed_prefix = None, out_prefix = None, family_id = False, overwrite = False, **kwargs):

    # Check if the ped prefix files exist
    if bed_prefix and confirm_bed_prefix(bed_prefix):

        # Assign the files
        bed_filename = bed_prefix + '.bed'
        bim_filename = bed_prefix + '.bim'
        fam_filename = bed_prefix + '.fam'

    # Check that all the bed files are assigned
    elif bed_filename and confirm_bed_files(bed_filename, bim_filename, fam_filename):
        
        # No action requried, just needs to pass
        pass 

    # Rename the fam file
    shutil.copy(fam_filename, fam_filename + '.pedind')

    # Update the fam filename
    fam_filename += '.pedind'

    # Assign the par filename
    par_filename = assign_unique_filename(out_prefix, 'par')

    # Assign the input files
    par_file_str = par_assign_input(bed_filename, bim_filename, fam_filename)

    # Add the Eigenstrat assignment lines
    par_file_str += par_assign_eigenstrat(out_prefix)

    # Add the family ID line
    par_file_str += par_assign_family_ids(family_id)

    # Create the par file
    par_file = open(par_filename, 'w')
    par_file.write(par_file_str)
    par_file.close()

    # Call convertf
    call_convertf(['-p', par_filename])

    # Remove the par file
    os.remove(par_filename)

    # Remove the copy
    os.remove(fam_filename)

    # Update the fam filename
    fam_filename = fam_filename[:-7]

def confirm_eigenstrat_files_from_prefix (eigenstrat_prefix):

    # Assign expected filenames
    geno_filename = eigenstrat_prefix + '.geno'
    ind_filename = eigenstrat_prefix + '.ind'
    snp_filename = eigenstrat_prefix + '.snp'

    # Check if a genotype file is assigned
    if not geno_filename:
        raise IOError('Unable to assign genotype file (i.e. --geno). Please confirm the file is named correctly')

    # Check if the ind and snp files are not assigned
    if not ind_filename and not snp_filename:
        raise IOError('Unable to assign the ind and snp files. Please confirm the files (i.e. --ind, --snp) are assigned')

    # Check if the ind is not assigned
    if not ind_filename:
        raise IOError('Unable to assign ind file. Please confirm the ind file (i.e. --ind) is assigned')

    # Check if the snp file is not assigned
    if not snp_filename:
        raise IOError('Unable to assign snp file. Please confirm the snp file (i.e. --snp) is assigned')

def confirm_eigenstrat_files (geno_filename, ind_filename, snp_filename):

    # Check if a genotype file is assigned
    if not geno_filename:
        raise IOError('Unable to assign genotype file (i.e. --geno). Please confirm the file is named correctly')

    # Check if the ind and snp files are not assigned
    if not ind_filename and not snp_filename:
        raise IOError('Unable to assign the ind and snp files. Please confirm the files (i.e. --ind, --snp) are assigned')

    # Check if the ind is not assigned
    if not ind_filename:
        raise IOError('Unable to assign ind file. Please confirm the ind file (i.e. --ind) is assigned')

    # Check if the snp file is not assigned
    if not snp_filename:
        raise IOError('Unable to assign snp file. Please confirm the snp file (i.e. --snp) is assigned')

def check_eigenstrat_prefix (geno_filename, ind_filename, snp_filename):

    # Assign the base name of each file
    geno_filename_wo_ext = os.path.splitext(geno_filename)[0]
    ind_basename_wo_ext = os.path.splitext(ind_filename)[0]
    snp_basename_wo_ext = os.path.splitext(snp_filename)[0]

    # Check if a prefix may be assigned without any changes
    if geno_filename_wo_ext == ind_basename_wo_ext == snp_basename_wo_ext:
        
        # Return True if a prefix may be assigned
        return True

    return False

def assign_eigenstrat_prefix (geno_filename, ind_filename, snp_filename):

    # Assign the base name of each file
    geno_filename_wo_ext = os.path.splitext(geno_filename)[0]

    # Check that the ind file needs to be changed
    if ind_filename != geno_filename_wo_ext + '.ind':

        # Rename the file
        shutil.move(ind_filename, geno_filename_wo_ext + '.ind')

    # Check that the ind file needs to be changed
    if snp_filename != geno_filename_wo_ext + '.snp':

        # Rename the file
        shutil.move(snp_filename, geno_filename_wo_ext + '.snp')

    # Return the prefix
    return geno_filename_wo_ext

def restore_eigenstrat_files (geno_filename, ind_filename, snp_filename):

    # Assign the base name of each file
    geno_filename_wo_ext = os.path.splitext(geno_filename)[0]

    # Check that the ind file needs to be changed
    if ind_filename != geno_filename_wo_ext + '.ind':

        # Rename the file
        shutil.move(geno_filename_wo_ext + '.ind', ind_filename)

    # Check that the ind file needs to be changed
    if snp_filename != geno_filename_wo_ext + '.snp':

        # Rename the file
        shutil.move(geno_filename_wo_ext + '.snp', snp_filename)

def assign_model_ind_filename (eigenstrat_prefix, model, overwrite = False):

    # Assign a model-based ind filename
    model_ind_filename = '%s.%s.ind' % (eigenstrat_prefix, model.name)

    # Check if the file should be overwritten
    if not overwrite:

        # Check if the output filename exists
        if os.path.isfile(model_ind_filename):
            raise Exception('Model-based individual file exists. Please use --overwrite to ignore.')

    return model_ind_filename

def create_ind_w_pops (eigenstrat_prefix, model, out_filename):

    # Create a temporary file to replace the ind file
    ind_pop_file = open(out_filename, 'w')

    # Open the ind file
    with open(eigenstrat_prefix + '.ind', 'r') as ind_file:

        # Loop the ind file
        for ind_line in ind_file:

            # Split the ind line
            ind_data = ind_line.strip().split()

            # Assign the current ind
            ind_name = ind_data[0]

            # Assign the pop from the ind
            pop_name = model.return_pop(ind_name)

            # Update the ind data
            ind_data[2] = pop_name
            
            # Write the updated line to the temporary file
            ind_pop_file.write(' '.join(ind_data) + '\n')

    ind_pop_file.close()

def update_ind_w_pops (eigenstrat_prefix, model):

    # Create a filename for the temporary ind file
    tmp_ind_filename = assign_unique_filename(eigenstrat_prefix + '.ind', 'tmp')

    # Create a temporary file that will replace the ind file
    ind_pop_file = open(tmp_ind_filename, 'w')

    # Open the ind file
    with open(eigenstrat_prefix + '.ind', 'r') as ind_file:

        # Loop the ind file
        for ind_line in ind_file:

            # Split the ind line
            ind_data = ind_line.strip().split()

            # Assign the current ind
            ind_name = ind_data[0]

            # Assign the pop from the ind
            pop_name = model.return_pop(ind_name)

            # Update the ind data
            ind_data[2] = pop_name
            
            # Write the updated line to the temporary file
            ind_pop_file.write(' '.join(ind_data) + '\n')

    ind_pop_file.close()

    # Rename the ind file
    shutil.move(tmp_ind_filename, eigenstrat_prefix + '.ind')

def r_admixr_d (eigenstrat_prefix, out_prefix, w_pops, x_pops, y_pops, z_pops):

    # Activate converions between pandas and R 
    pandas2ri.activate()

    # Activate the admixr library
    try:
        r_admixr_lib = importr('admixr')
    except:
        raise Exception('Please install the admixr R library')

    # Create the c() R object
    rc = robjects.r['c']

    # Create the eigenstrat file object from admixr
    r_eigenstrat = robjects.r.eigenstrat

    # Create the statistic objects from admixr
    r_d = robjects.r.d
    
    # Load the eigenstrat file into R
    r_eigenstrat_input = r_eigenstrat(eigenstrat_prefix)

    # Run the D statistic
    r_d_results = r_d(W = rc(w_pops), X = rc(x_pops), Y = rc(y_pops), Z = rc(z_pops), data = r_eigenstrat_input)

    # Convert the results into a Pandas dataframe
    d_results = pandas2ri.ri2py(r_d_results)

    # Change the datatype of the nsnps, ABBA, and BABA columns
    d_results[['nsnps', 'BABA', 'ABBA']] = d_results[['nsnps', 'BABA', 'ABBA']].astype(int)

    # Write the dataframe as a file
    d_results.to_csv(out_prefix + '.d', sep = '\t', index = False)

def r_admixr_f4 (eigenstrat_prefix, out_prefix, w_pops, x_pops, y_pops, z_pops):

    # Activate converions between pandas and R 
    pandas2ri.activate()

    # Activate the admixr library
    try:
        r_admixr_lib = importr('admixr')
    except:
        raise Exception('Please install the admixr R library')

    # Create the c() R object
    rc = robjects.r['c']

    # Create the eigenstrat file object from admixr
    r_eigenstrat = robjects.r.eigenstrat

    # Create the statistic objects from admixr
    r_f4 = robjects.r.f4
    
    # Load the eigenstrat file into R
    r_eigenstrat_input = r_eigenstrat(eigenstrat_prefix)

    # Run the F4 statistic
    r_f4_results = r_f4(W = rc(w_pops), X = rc(x_pops), Y = rc(y_pops), Z = rc(z_pops), data = r_eigenstrat_input)

    # Convert the results into a Pandas dataframe
    f4_results = pandas2ri.ri2py(r_f4_results)

    # Change the datatype of the nsnps, ABBA, and BABA columns
    f4_results[['nsnps', 'BABA', 'ABBA']] = f4_results[['nsnps', 'BABA', 'ABBA']].astype(int)

    # Write the dataframe as a file
    f4_results.to_csv(out_prefix + '.f4', sep = '\t', index = False)

def r_admixr_f4ratio (eigenstrat_prefix, out_prefix, a_pops, b_pops, c_pops, x_pops, o_pops):

    # Activate converions between pandas and R 
    pandas2ri.activate()

    # Activate the admixr library
    try:
        r_admixr_lib = importr('admixr')
    except:
        raise Exception('Please install the admixr R library')

    # Create the c() R object
    rc = robjects.r['c']

    # Create the eigenstrat file object from admixr
    r_eigenstrat = robjects.r.eigenstrat

    # Create the statistic objects from admixr
    r_f4ratio = robjects.r.f4ratio
    
    # Load the eigenstrat file into R
    r_eigenstrat_input = r_eigenstrat(eigenstrat_prefix)

    # Run the f4ratio statistic
    r_f4ratio_results = r_f4ratio(A = rc(a_pops), B = rc(b_pops), C = rc(c_pops), X = rc(x_pops), O = rc(o_pops), data = r_eigenstrat_input)

    # Convert the results into a Pandas dataframe
    f4ratio_results = pandas2ri.ri2py(r_f4ratio_results)

    # Write the dataframe as a file
    f4ratio_results.to_csv(out_prefix + '.f4ratio', sep = '\t', index = False)

def r_admixr_f3 (eigenstrat_prefix, out_prefix, a_pops, b_pops, c_pops):

    # Activate converions between pandas and R 
    pandas2ri.activate()

    # Activate the admixr library
    try:
        r_admixr_lib = importr('admixr')
    except:
        raise Exception('Please install the admixr R library')

    # Create the c() R object
    rc = robjects.r['c']

    # Create the eigenstrat file object from admixr
    r_eigenstrat = robjects.r.eigenstrat

    # Create the statistic objects from admixr
    r_f3 = robjects.r.f3
    
    # Load the eigenstrat file into R
    r_eigenstrat_input = r_eigenstrat(eigenstrat_prefix)

    # Run the f3 statistic
    r_f3_results = r_f3(A = rc(a_pops), B = rc(b_pops), C = rc(c_pops), data = r_eigenstrat_input)

    # Convert the results into a Pandas dataframe
    f3_results = pandas2ri.ri2py(r_f3_results)

    # Change the datatype of the nsnps column
    f3_results['nsnps'] = f3_results['nsnps'].astype(int)

    # Write the dataframe as a file
    f3_results.to_csv(out_prefix + '.f3', sep = '\t', index = False)


import pysam
import argparse
import sys
from pgpipe.logging_module import initLogger
from pgpipe.model import read_model_file
from pgpipe.vcf_reader_func import VcfReader


def createParser():
    parser = argparse.ArgumentParser(description=("Remove individuals from a single VCF file"))
    parser.add_argument("--vcf",dest="vcfname",help=("VCF input filename"))
    parser.add_argument("--model-file",dest="modelname",help=("Model filename"))
    parser.add_argument("--model",dest="poptag",help=("Name of model if multiple in model file"))
    parser.add_argument("--out",dest="outname",default="-",help=("Name of VCF output file"))
    return parser






def filter_single_vcf(sysargs):
    parser = createParser()
    args = parser.parse_args(sysargs)
    popmodel = None
    if args.modelname is not None:
        popmodels = read_model_file(args.modelname)
        if len(popmodels) != 1:
            popmodel = popmodels[args.poptag]
        else:
            pp = list(popmodels.keys())
            popmodel = popmodels[pp[0]]
    vcf_in = VcfReader(args.vcfname,
                       popmodel=popmodel)
    sample_names = [l for l in vcf_in.reader.header.samples]
    bad_names = [False for l in sample_names]
    for rec in vcf_in.reader:
        #for i,samp in enumerate(rec.samples):
        for i in range(len(rec.samples)):
            samp = rec.samples[i]
            if samp.alleles[0] in [None,'N']:
                bad_names[i] = True
    final_names = []
    for i in range(len(sample_names)):
        if not bad_names[i]:
            final_names.append(sample_names[i])

    vcf_in.close()
    sys.stderr.write(str(len(final_names))+'\n')
    vcf_in = VcfReader(args.vcfname)
    vcf_in.reader.subset_samples(final_names)
    vcf_out = pysam.VariantFile(args.outname,'w',header=vcf_in.reader.header)
    rec_list = vcf_in.getRecordList()
    for rec in rec_list:
        vcf_out.write(rec)
    vcf_in.close()
    vcf_out.close()
    


if __name__ == "__main__":
    initLogger()
    filter_single_vcf(sys.argv[1:])


#!/usr/bin/env python
'''
    Utilites for BED-formatted files

    Automates various utilites for BED-formatted files. This currently includes:
    MORE INFO.
    
    ############################
    Input Command-line Arguments
    ############################
    **--bed** *<input_filename>*
        Argument used to define the filename of the BED file.
    **--beds** *<input_filename>* *<input1_filename, input2_filename, etc.>*
        Argument used to define the filename of the BED file(s). May be used multiple 
        times.
    **--chrom-file** *<chrom_filename>*
        Argument used to define the filename of a file with the sizes of each
        chromosome file. Appropriate files may be downloaded from the UCSC Genome
        Browser.
    
    #############################
    Output Command-line Arguments
    #############################
    **--out** *<output_filename>*
        Argument used to define the complete output filename.
    **--overwrite**
        Argument used to define if previous output should be overwritten.
    
    ##################################
    Utility Command-line Specification
    ##################################
    **--utility** *<subtract, extend, sort, merge, complement>*
        Argument used to define the desired utility. Current utilities include: subtract
        features from a BED file that overlap with features within a second BED file 
        (subtract); extend the flanks of features upstream, downstream, or both within a 
        single BED file (extend); sort the features within a single BED file (sort); merge
        features within one or more BED files (merge); create a BED file of complementary
        features - i.e. features that do not overlap - from a BED file (complement).

    ***************************************
    Subtract Utility Command-line Arguments
    ***************************************
    **--subtract-bed** *<subtract_file_filename>*
        Argument used to define the BED file used for removing features/positions.
    **--subtract-entire-feature**
        Argument used to define if entire features within the input BED should be 
        removed if they overlap with features in subtract-bed.
    **--min-reciprocal-overlap** *<overlap_float>*
        Argument used to define the minimum reciprocal overlap of features required 
        for removal (e.g. 0.1 indicates 10% overlap).
    **--min-input-overlap** *<overlap_float>*
        Argument used to define the minimum overlap of input features required for 
        removal.
    **--min-subtract-overlap** *<overlap_float>*
        Argument used to define the minimum overlap of subtract-bed features required 
        for removal.
    **--subtract-entire-feature**
        Argument used to define that features should be removed from the input BED if
        the minimum overlap of **--min-input-overlap** or **--min-subtract-overlap** 
        is reached.

    *************************************
    Extend Utility Command-line Arguments
    *************************************
    **--extend-flanks** *<bp_int>*
        Argument used to define the length of base pairs (bp) to extend both upstream
        and downstream of features.
    **--extend-upstream** *<bp_int>*
        Argument used to define the length of base pairs (bp) to extend upstream of 
        features.
    **--extend-downstream** *<bp_int>*
        Argument used to define the length of base pairs (bp) to extend downstream of 
        features.

    ************************************
    Merge Utility Command-line Arguments
    ************************************
    **--max-merge-distance** *<bp_int>*
        Argument used to define the maximum distance allowed between features to be 
        merged.
'''

import os
import sys
import subprocess
import shutil
import argparse
import glob
import logging

#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir, 'pppipe')))

# Import basic bedtools functions
from pgpipe.bedtools import merge_bed_files, standard_bedtools_call

from pgpipe.logging_module import initLogger, logArgs

def bed_argument_parser(passed_arguments):
    '''Phase Argument Parser - Assigns arguments for vcftools from command line.
    Depending on the argument in question, a default value may be specified'''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError # File not found
                setattr(args, self.dest, value)
        return customAction

    def parser_confirm_file_list ():
        '''Custom action to confirm file exists in list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                # Loop the list
                for value_item in value:
                    # Check if the file exists
                    if not os.path.isfile(value_item):
                        raise IOError('%s not found' % value_item)
                if not getattr(args, self.dest):
                    setattr(args, self.dest, value)
                else:
                    getattr(args, self.dest).extend(value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    bed_parser = argparse.ArgumentParser()

    # Input arguments
    bed_input = bed_parser.add_mutually_exclusive_group(required = True)
    bed_input.add_argument('--bed', help = 'Defines the filename of the BED file', type = str, action = parser_confirm_file())
    bed_input.add_argument('--beds', help = 'Defines the filenames of the BED files (may be used multiple times)', type = str, nargs = '+', action = parser_confirm_file_list())
    bed_parser.add_argument('--chrom-file', help = 'File of chromosome sizes', type = str, action = parser_confirm_file())

    # Utility based arguments
    utility_list = ['subtract', 'extend', 'sort', 'merge', 'complement']
    bed_parser.add_argument('--utility', metavar = metavar_list(utility_list), help = 'Specifies the utility to be used', type = str, choices = utility_list, required = True)

    # Subtract-specific arguments
    subtract_group = bed_parser.add_argument_group('Subtract Utility Arguments')
    subtract_group.add_argument('--subtract-bed', help = 'BED file used for removing features/positions', type = str, action = parser_confirm_file())
    subtract_group.add_argument('--subtract-entire-feature', help = 'Remove entire features within the input if they overlap with features in subtract-bed', action = 'store_true')
    subtract_group.add_argument('--min-reciprocal-overlap', help = 'Minimum reciprocal overlap of features required for removal', type = float)
    subtract_group.add_argument('--min-input-overlap', help = 'Minimum overlap of input features required for removal', type = float)
    subtract_group.add_argument('--min-subtract-overlap', help = 'Minimum overlap of subtract-bed features required for removal', type = float)
    subtract_group.add_argument('--either-min-overlap', help = 'Remove a feature if either minimum overlap is reached ', action = 'store_true')

    # Extend-specific arguments
    extend_group = bed_parser.add_argument_group('Extend Utility Arguments')
    extend_group.add_argument('--extend-flanks', help = 'Base pairs to extend both flanks', type = int)
    extend_group.add_argument('--extend-upstream', help = 'Base pairs to extend upstream', type = int)
    extend_group.add_argument('--extend-downstream', help = 'Base pairs to extend downstream', type = int)

    # Extend-specific arguments
    merge_group = bed_parser.add_argument_group('Merge Utility Arguments')
    merge_group.add_argument('--max-merge-distance', help = 'Max distance allowed between features to be merged', type = int)

    # Other basic arguments. Expand as needed
    bed_parser.add_argument('--out', help = 'Defines the output filename', default = 'out.bed')
    bed_parser.add_argument('--overwrite', help = "Defines that previous output files should be overwritten", action = 'store_true')

    if passed_arguments:
        return bed_parser.parse_args(passed_arguments)
    else:
        return bed_parser.parse_args()

def run (passed_arguments = []):
    '''
        Utilites for BED-formatted files

        This function uses the argparse-based function :py:func:`bed_argument_parser`
        to parse either sys.argv or passed_arguments to obtain the parameters below. 
        The parameters are then translated to their BEDtools equivalent. Once all the 
        parameters are assigned, BEDtools is called.

        Parameters
        ----------
        --bed : str
            Filename of the BED
        --beds : list
            List of BED filenames
        --chrom-file : str
            File of chromosome sizes
        --utility : str
            Utility to be used
        --out : str
            Complete output filename
        --overwrite
            Species if previous output should be overwriten
        --subtract-bed : str
            Filename of the BED to subtract
        --subtract-entire-feature : bool
            Remove entire feature if there is an overlap in --subtract-bed
        --min-reciprocal-overlap : float
            Minimum reciprocal overlap for feature removal
        --min-input-overlap : float
            Minimum input (i.e. --bed) overlap for feature removal
        --min-subtract-overlap : float
            Minimum --subtract-bed overlap for feature removal
        --either-min-overlap : bool
            Remove a feature if either minimum overlap is reached
        --extend-flanks : int
            Base pairs to extend both flanks
        --extend-upstream : int
            Base pairs to extend upstream
        --extend-downstream : int
            Base pairs to extend downstream
        --max-merge-distance : int
            Max distance allowed between features to be merged

        Raises
        ------
        IOError
            Output file already exists
        Exception
            More than one BED given - unless using the merge utility
        Exception
            No extension arguments given with extend utility
        Exception
            No --chrom-file given with extend/complement utilities

    '''

    # Grab BED arguments from command line
    bed_args = bed_argument_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(bed_args, 'bed_utilities')

    # Check if previous output should not be overwriten
    if not bed_args.overwrite:

        # Check if previous output exists
        if os.path.exists(bed_args.out):
            raise IOError('Utility output already exists')

    # Check if the utility only functions with a single BED file
    if bed_args.beds and bed_args.utility not in ['merge']:
        raise Exception('%s only supports a single BED file' % bed_args.utility)

    # Check if the user has selected the merge utility
    if bed_args.utility == 'merge':

        # Create list of BED files to merge
        bed_files_to_merge = []

        # Check if multiple BED files have been specified
        if bed_args.beds:

            # Assign the specified BED files 
            bed_files_to_merge = bed_args.beds

        # Check if a single BED file has been specified
        if bed_args.bed:

            # Assign the specified BED files 
            bed_files_to_merge = [bed_args.bed]

        # Create list of optional merge args
        optional_merge_args = []

        # Check if --max-merge-distance has been specified 
        if bed_args.max_merge_distance:
            optional_merge_args.extend(['-d', bed_args.max_merge_distance])

        # Merge the BED files
        merge_bed_files (bed_files_to_merge, bed_args.out, optional_merge_args)

    # Catch the rest of the utilites 
    else:

        # Check if the utility requires a chromosome size file
        if bed_args.utility in ['complement', 'extend']:

            # Check if a chromosome sizes file has been defined
            if not bed_args.chrom_file:
                raise Exception('%s requires a file of chromosome sizes. Please use --chrom-file' % bed_args.utility)

        # Create list of bedtools
        bedtools_arg_list = []
        
        # Check if the subtract utility has been selected
        if bed_args.utility == 'subtract':

            # Check that a subtract BED file has been specified
            if not bed_args.subtract_bed:
                raise Exception('%s requires a subtract BED file. Please use --subtract-bed' % bed_args.utility)

            # Assign the utility
            bedtools_arg_list.append(bed_args.utility)

            # Assign the input arguments
            bedtools_arg_list.extend(['-a', bed_args.bed])

            # Assign the --subtract-bed argument
            bedtools_arg_list.extend(['-b', bed_args.subtract_bed])

            # Check if --subtract-entire-feature has been specified 
            if bed_args.subtract_entire_feature:
                bedtools_arg_list.append('-A')

            # Check if --min-reciprocal-overlap has been specified 
            if bed_args.min_reciprocal_overlap:
                bedtools_arg_list.extend(['-f', bed_args.min_reciprocal_overlap, '-r'])

            # Check if --min-input-overlap has been specified 
            if bed_args.min_input_overlap:
                bedtools_arg_list.extend(['-f', bed_args.min_input_overlap])

            # Check if --min-subtract-overlap has been specified 
            if bed_args.min_subtract_overlap:
                bedtools_arg_list.extend(['-F', bed_args.min_subtract_overlap])

            # Check if --either-min-overlap has been specified 
            if bed_args.either_min_overlap:
                bedtools_arg_list.append('-e')

        # Check if the extend utility has been selected
        elif bed_args.utility == 'extend':

            # Confirm at least one extension argument has been assigned
            if not bed_args.extend_flanks and not bed_args.extend_upstream and not bed_args.extend_downstream:
                raise Exception('%s requires at least one extension argument (e.g. --extend-flanks)' % bed_args.utility)

            # Assign the utility
            bedtools_arg_list.append('slop')

            # Assign the input arguments
            bedtools_arg_list.extend(['-i', bed_args.bed])

            # Assign the --chrom-file argument
            bedtools_arg_list.extend(['-g', bed_args.chrom_file])

            # Check if --extend-flanks has been specified 
            if bed_args.extend_flanks:
                bedtools_arg_list.extend(['-b', bed_args.extend_flanks])

            # Check if --extend-upstream has been specified 
            if bed_args.extend_upstream:
                bedtools_arg_list.extend(['-l', bed_args.extend_upstream])

            # Check if --extend-downstream has been specified 
            if bed_args.extend_downstream:
                bedtools_arg_list.extend(['-r', bed_args.extend_downstream])

        # Check if the sort utility has been selected
        elif bed_args.utility == 'sort':

            # Assign the utility
            bedtools_arg_list.append(bed_args.utility)

            # Assign the input arguments
            bedtools_arg_list.extend(['-i', bed_args.bed])

        # Check if the complement utility has been selected
        elif bed_args.utility == 'complement':

            # Assign the utility
            bedtools_arg_list.append(bed_args.utility)

            # Assign the input arguments
            bedtools_arg_list.extend(['-i', bed_args.bed])

            # Assign the --chrom-file argument
            bedtools_arg_list.extend(['-g', bed_args.chrom_file])

        # Catch unknown utilities    
        else:
            raise Exception('%s is an unknown utility. Please check input' % bed_args.utility)

        # Call bedtools
        standard_bedtools_call(bedtools_arg_list, bed_args.out)

if __name__ == "__main__":
    initLogger()
    run()



import sys
import logging



def initLogger(filename='pipeline.log', filelevel='INFO',
               streamlevel='WARNING', resetlog=True):
    """General logger initialization for PPP functions.

    Messages from WARNING level and higher will be logged to file and to
    stderr. By default, INFO level will also be written to logfile. Both
    levels are flexible.

    Level groupings:
        ERROR: Error messages should be generated by an exception call
        WARNING: Non-terminal behavior that may be unusual (i.e. lists
            with no values, blank strings)
        INFO: Every major (user-facing) function should have the following:
            -Message for function start
            -List of input arguments and options
            -Basic sanity checks (dimensions of input data)
            -Statements before or after major function calls
            -Message for function end
        DEBUG: Mainly for developer use/debugging. Generate logs for
            sub-functions that match INFO level for major functions.
            Possible care should be used if there are lots of loops called.

    Use: Call with either the individual function (in __name__=="__main__"
    statement) or in pipeline file.

    Parameters
    ----------
    filename : str ("pipeline.log")
        Name of file that log will be written to

    filelevel : {'INFO','DEBUG','WARNING','ERROR'}
        Set minimum level of log messages that are written to log file.
        Note that this acts as a de facto minumum for 'streamlevel' as well.

    streamlevel : {'WARNING','DEBUG','INFO','ERROR'}
        Set minimum level of log messages that are output to stream.

    resetlog : bool (True)
        If true, will overwrite logfile when opening. Set to false if log is
        being initialized multiple times

    Returns
    -------
    None

    Exceptions
    ----------
    If filelevel or streamlevel are not a valid logging level

    """
    log_levels = ['DEBUG','INFO','WARNING','ERROR']
    if filelevel is not None and filelevel.upper() not in log_levels:
        raise Exception('filelevel value %s is not a valid level' %
                         filelevel)
    if streamlevel is not None and streamlevel.upper() not in log_levels:
        raise Exception('streamlevel value %s is not a valid level' %
                         streamlevel)
    fmt_def = "%(asctime)s - %(funcName)s - %(levelname)s: %(message)s"
    fmt_notime = "%(funcName)s - %(levelname)s: %(message)s"
    fmtr = logging.Formatter(fmt=fmt_def)
    fmtr_notime = logging.Formatter(fmt=fmt_notime)
    filelogger = logging.getLogger()
    filelogger.setLevel('INFO')
    if streamlevel is not None:
        s_handler = logging.StreamHandler()
        s_handler.setFormatter(fmtr_notime)
        s_handler.setLevel(streamlevel)
        filelogger.addHandler(s_handler)
    logmode = 'a'
    if resetlog:
        logmode = 'w'
    if filelevel is not None:
        f_handler = logging.FileHandler(filename,mode=logmode)
        f_handler.setFormatter(fmtr)
        f_handler.setLevel(filelevel)
        #filelogger.setLevel(filelevel)
        filelogger.addHandler(f_handler)
    #Formats exception messages to be sent to appropriate loggers
    def exp_handler(etype,val,tb):
        logging.error("%s" % (val), exc_info=(etype,val,tb))

    sys.excepthook = exp_handler


def logArgs(args, func_name=None, print_nones=False):
    header = "Arguments"
    if func_name is not None:
        header+=" for"+func_name
    for arg in vars(args):
        val = vars(args)[arg]
        if val is not None or print_nones:
            logging.info('Argument %s: %s' % (arg,val))
